from __future__ import unicode_literals

import json
import logging
from builtins import str
from collections import defaultdict
from itertools import combinations
from pathlib import Path

from future.utils import iteritems, itervalues
from snips_nlu_utils import normalize, hash_str

from snips_nlu.common.log_utils import log_elapsed_time, log_result
from snips_nlu.common.utils import (
    check_persisted_path, deduplicate_overlapping_entities, fitted_required,
    json_string)
from snips_nlu.constants import (
    DATA, END, ENTITIES, ENTITY, ENTITY_KIND, INTENTS, LANGUAGE, RES_INTENT,
    RES_INTENT_NAME, RES_MATCH_RANGE, RES_SLOTS, SLOT_NAME, START, TEXT,
    UTTERANCES, RES_PROBA)
from snips_nlu.dataset import (
    validate_and_format_dataset, extract_intent_entities)
from snips_nlu.dataset.utils import get_stop_words_whitelist
from snips_nlu.entity_parser.builtin_entity_parser import is_builtin_entity
from snips_nlu.exceptions import IntentNotFoundError, LoadingError
from snips_nlu.intent_parser.intent_parser import IntentParser
from snips_nlu.pipeline.configs import LookupIntentParserConfig
from snips_nlu.preprocessing import tokenize_light
from snips_nlu.resources import get_stop_words
from snips_nlu.result import (
    empty_result, intent_classification_result, parsing_result,
    unresolved_slot, extraction_result)

logger = logging.getLogger(__name__)


@IntentParser.register("lookup_intent_parser")
class LookupIntentParser(IntentParser):
    """A deterministic Intent parser implementation based on a dictionary

    This intent parser is very strict by nature, and tends to have a very good
    precision but a low recall. For this reason, it is interesting to use it
    first before potentially falling back to another parser.
    """

    config_type = LookupIntentParserConfig

    def __init__(self, config=None, **shared):
        """The lookup intent parser can be configured by passing a
        :class:`.LookupIntentParserConfig`"""
        super(LookupIntentParser, self).__init__(config, **shared)
        self._language = None
        self._stop_words = None
        self._stop_words_whitelist = None
        self._map = None
        self._intents_names = []
        self._slots_names = []
        self._intents_mapping = dict()
        self._slots_mapping = dict()
        self._entity_scopes = None

    @property
    def language(self):
        return self._language

    @language.setter
    def language(self, value):
        self._language = value
        if value is None:
            self._stop_words = None
        else:
            if self.config.ignore_stop_words:
                self._stop_words = get_stop_words(self.resources)
            else:
                self._stop_words = set()

    @property
    def fitted(self):
        """Whether or not the intent parser has already been trained"""
        return self._map is not None

    @log_elapsed_time(
        logger, logging.INFO, "Fitted lookup intent parser in {elapsed_time}")
    def fit(self, dataset, force_retrain=True):
        """Fits the intent parser with a valid Snips dataset"""
        logger.info("Fitting lookup intent parser...")
        dataset = validate_and_format_dataset(dataset)
        self.load_resources_if_needed(dataset[LANGUAGE])
        self.fit_builtin_entity_parser_if_needed(dataset)
        self.fit_custom_entity_parser_if_needed(dataset)
        self.language = dataset[LANGUAGE]
        self._entity_scopes = _get_entity_scopes(dataset)
        self._map = dict()
        self._stop_words_whitelist = get_stop_words_whitelist(
            dataset, self._stop_words)
        entity_placeholders = _get_entity_placeholders(dataset, self.language)

        ambiguous_keys = set()
        for (key, val) in self._generate_io_mapping(dataset[INTENTS],
                                                    entity_placeholders):
            key = hash_str(key)
            # handle key collisions -*- flag ambiguous entries -*-
            if key in self._map and self._map[key] != val:
                ambiguous_keys.add(key)
            else:
                self._map[key] = val

        # delete ambiguous keys
        for key in ambiguous_keys:
            self._map.pop(key)

        return self

    @log_result(logger, logging.DEBUG, "LookupIntentParser result -> {result}")
    @log_elapsed_time(logger, logging.DEBUG, "Parsed in {elapsed_time}.")
    @fitted_required
    def parse(self, text, intents=None, top_n=None):
        """Performs intent parsing on the provided *text*

        Intent and slots are extracted simultaneously through pattern matching

        Args:
            text (str): input
            intents (str or list of str): if provided, reduces the scope of
                intent parsing to the provided list of intents
            top_n (int, optional): when provided, this method will return a
                list of at most top_n most likely intents, instead of a single
                parsing result.
                Note that the returned list can contain less than ``top_n``
                elements, for instance when the parameter ``intents`` is not
                None, or when ``top_n`` is greater than the total number of
                intents.

        Returns:
            dict or list: the most likely intent(s) along with the extracted
            slots. See :func:`.parsing_result` and :func:`.extraction_result`
            for the output format.

        Raises:
            NotTrained: when the intent parser is not fitted
        """
        if top_n is None:
            top_intents = self._parse_top_intents(text, top_n=1,
                                                  intents=intents)
            if top_intents:
                intent = top_intents[0][RES_INTENT]
                slots = top_intents[0][RES_SLOTS]
                if intent[RES_PROBA] <= 0.5:
                    # return None in case of ambiguity
                    return empty_result(text, probability=1.0)
                return parsing_result(text, intent, slots)
            return empty_result(text, probability=1.0)
        return self._parse_top_intents(text, top_n=top_n, intents=intents)

    def _parse_top_intents(self, text, top_n, intents=None):
        if isinstance(intents, str):
            intents = {intents}
        elif isinstance(intents, list):
            intents = set(intents)

        if top_n < 1:
            raise ValueError(
                "top_n argument must be greater or equal to 1, but got: %s"
                % top_n)

        results_per_intent = defaultdict(list)
        for text_candidate, entities in self._get_candidates(text, intents):
            val = self._map.get(hash_str(text_candidate))
            if val is not None:
                result = self._parse_map_output(text, val, entities, intents)
                if result:
                    intent_name = result[RES_INTENT][RES_INTENT_NAME]
                    results_per_intent[intent_name].append(result)

        results = []
        for intent_results in itervalues(results_per_intent):
            sorted_results = sorted(intent_results,
                                    key=lambda res: len(res[RES_SLOTS]))
            results.append(sorted_results[0])

        # In some rare cases there can be multiple ambiguous intents
        # In such cases, priority is given to results containing fewer slots
        weights = [1.0 / (1.0 + len(res[RES_SLOTS])) for res in results]
        total_weight = sum(weights)

        for res, weight in zip(results, weights):
            res[RES_INTENT][RES_PROBA] = weight / total_weight

        results = sorted(results, key=lambda r: -r[RES_INTENT][RES_PROBA])
        return results[:top_n]

    def _get_candidates(self, text, intents):
        candidates = defaultdict(list)
        for grouped_entity_scope in self._entity_scopes:
            entity_scope = grouped_entity_scope["entity_scope"]
            intent_group = grouped_entity_scope["intent_group"]
            intent_group = [intent_ for intent_ in intent_group
                            if intents is None or intent_ in intents]
            if not intent_group:
                continue

            builtin_entities = self.builtin_entity_parser.parse(
                text, scope=entity_scope["builtin"], use_cache=True)
            custom_entities = self.custom_entity_parser.parse(
                text, scope=entity_scope["custom"], use_cache=True)
            all_entities = builtin_entities + custom_entities
            all_entities = deduplicate_overlapping_entities(all_entities)

            # We generate all subsets of entities to match utterances
            # containing ambivalent words which can be both entity values or
            # random words
            for entities in _get_entities_combinations(all_entities):
                processed_text = self._replace_entities_with_placeholders(
                    text, entities)
                for intent in intent_group:
                    cleaned_text = self._preprocess_text(text, intent)
                    cleaned_processed_text = self._preprocess_text(
                        processed_text, intent)

                    raw_candidate = cleaned_text, []
                    placeholder_candidate = cleaned_processed_text, entities
                    intent_candidates = [raw_candidate, placeholder_candidate]
                    for text_input, text_entities in intent_candidates:
                        if text_input not in candidates \
                                or text_entities not in candidates[text_input]:
                            candidates[text_input].append(text_entities)
                            yield text_input, text_entities

    def _parse_map_output(self, text, output, entities, intents):
        """Parse the map output to the parser's result format"""
        intent_id, slot_ids = output
        intent_name = self._intents_names[intent_id]
        if intents is not None and intent_name not in intents:
            return None

        parsed_intent = intent_classification_result(
            intent_name=intent_name, probability=1.0)
        slots = []
        # assert invariant
        assert len(slot_ids) == len(entities)
        for slot_id, entity in zip(slot_ids, entities):
            slot_name = self._slots_names[slot_id]
            rng_start = entity[RES_MATCH_RANGE][START]
            rng_end = entity[RES_MATCH_RANGE][END]
            slot_value = text[rng_start:rng_end]
            entity_name = entity[ENTITY_KIND]
            slot = unresolved_slot(
                [rng_start, rng_end], slot_value, entity_name, slot_name)
            slots.append(slot)

        return extraction_result(parsed_intent, slots)

    @fitted_required
    def get_intents(self, text):
        """Returns the list of intents ordered by decreasing probability

        The length of the returned list is exactly the number of intents in the
        dataset + 1 for the None intent
        """
        nb_intents = len(self._intents_names)
        top_intents = [intent_result[RES_INTENT] for intent_result in
                       self._parse_top_intents(text, top_n=nb_intents)]
        matched_intents = {res[RES_INTENT_NAME] for res in top_intents}
        for intent in self._intents_names:
            if intent not in matched_intents:
                top_intents.append(intent_classification_result(intent, 0.0))

        # The None intent is not included in the lookup table and is thus
        # never matched by the lookup parser
        top_intents.append(intent_classification_result(None, 0.0))
        return top_intents

    @fitted_required
    def get_slots(self, text, intent):
        """Extracts slots from a text input, with the knowledge of the intent

        Args:
            text (str): input
            intent (str): the intent which the input corresponds to

        Returns:
            list: the list of extracted slots

        Raises:
            IntentNotFoundError: When the intent was not part of the training
                data
        """
        if intent is None:
            return []

        if intent not in self._intents_names:
            raise IntentNotFoundError(intent)

        slots = self.parse(text, intents=[intent])[RES_SLOTS]
        if slots is None:
            slots = []
        return slots

    def _get_intent_stop_words(self, intent):
        whitelist = self._stop_words_whitelist.get(intent, set())
        return self._stop_words.difference(whitelist)

    def _get_intent_id(self, intent_name):
        """generate a numeric id for an intent

        Args:
            intent_name (str): intent name

        Returns:
            int: numeric id

        """
        intent_id = self._intents_mapping.get(intent_name)
        if intent_id is None:
            intent_id = len(self._intents_names)
            self._intents_names.append(intent_name)
            self._intents_mapping[intent_name] = intent_id

        return intent_id

    def _get_slot_id(self, slot_name):
        """generate a numeric id for a slot

        Args:
            slot_name (str): intent name

        Returns:
            int: numeric id

        """
        slot_id = self._slots_mapping.get(slot_name)
        if slot_id is None:
            slot_id = len(self._slots_names)
            self._slots_names.append(slot_name)
            self._slots_mapping[slot_name] = slot_id

        return slot_id

    def _preprocess_text(self, txt, intent):
        """Replaces stop words and characters that are tokenized out by
            whitespaces"""
        stop_words = self._get_intent_stop_words(intent)
        tokens = tokenize_light(txt, self.language)
        cleaned_string = " ".join(
            [tkn for tkn in tokens if normalize(tkn) not in stop_words])
        return cleaned_string.lower()

    def _generate_io_mapping(self, intents, entity_placeholders):
        """Generate input-output pairs"""
        for intent_name, intent in sorted(iteritems(intents)):
            intent_id = self._get_intent_id(intent_name)
            for entry in intent[UTTERANCES]:
                yield self._build_io_mapping(
                    intent_id, entry, entity_placeholders)

    def _build_io_mapping(self, intent_id, utterance, entity_placeholders):
        input_ = []
        output = [intent_id]
        slots = []
        for chunk in utterance[DATA]:
            if SLOT_NAME in chunk:
                slot_name = chunk[SLOT_NAME]
                slot_id = self._get_slot_id(slot_name)
                entity_name = chunk[ENTITY]
                placeholder = entity_placeholders[entity_name]
                input_.append(placeholder)
                slots.append(slot_id)
            else:
                input_.append(chunk[TEXT])
        output.append(slots)

        intent = self._intents_names[intent_id]
        key = self._preprocess_text(" ".join(input_), intent)

        return key, output

    def _replace_entities_with_placeholders(self, text, entities):
        if not entities:
            return text
        entities = sorted(entities, key=lambda e: e[RES_MATCH_RANGE][START])
        processed_text = ""
        current_idx = 0
        for ent in entities:
            start = ent[RES_MATCH_RANGE][START]
            end = ent[RES_MATCH_RANGE][END]
            processed_text += text[current_idx:start]
            place_holder = _get_entity_name_placeholder(
                ent[ENTITY_KIND], self.language)
            processed_text += place_holder
            current_idx = end
        processed_text += text[current_idx:]

        return processed_text

    @check_persisted_path
    def persist(self, path):
        """Persists the object at the given path"""
        path.mkdir()
        parser_json = json_string(self.to_dict())
        parser_path = path / "intent_parser.json"

        with parser_path.open(mode="w", encoding="utf8") as pfile:
            pfile.write(parser_json)
        self.persist_metadata(path)

    @classmethod
    def from_path(cls, path, **shared):
        """Loads a :class:`LookupIntentParser` instance from a path

        The data at the given path must have been generated using
        :func:`~LookupIntentParser.persist`
        """
        path = Path(path)
        model_path = path / "intent_parser.json"
        if not model_path.exists():
            raise LoadingError(
                "Missing lookup intent parser metadata file: %s"
                % model_path.name)

        with model_path.open(encoding="utf8") as pfile:
            metadata = json.load(pfile)
        return cls.from_dict(metadata, **shared)

    def to_dict(self):
        """Returns a json-serializable dict"""
        stop_words_whitelist = None
        if self._stop_words_whitelist is not None:
            stop_words_whitelist = {
                intent: sorted(values)
                for intent, values in iteritems(self._stop_words_whitelist)}
        return {
            "config": self.config.to_dict(),
            "language_code": self.language,
            "map": self._map,
            "slots_names": self._slots_names,
            "intents_names": self._intents_names,
            "entity_scopes": self._entity_scopes,
            "stop_words_whitelist": stop_words_whitelist,
        }

    @classmethod
    def from_dict(cls, unit_dict, **shared):
        """Creates a :class:`LookupIntentParser` instance from a dict

        The dict must have been generated with
        :func:`~LookupIntentParser.to_dict`
        """
        config = cls.config_type.from_dict(unit_dict["config"])
        parser = cls(config=config, **shared)
        parser.language = unit_dict["language_code"]
        # pylint:disable=protected-access
        parser._map = _convert_dict_keys_to_int(unit_dict["map"])
        parser._slots_names = unit_dict["slots_names"]
        parser._intents_names = unit_dict["intents_names"]
        parser._entity_scopes = unit_dict["entity_scopes"]
        if parser.fitted:
            whitelist = unit_dict["stop_words_whitelist"]
            parser._stop_words_whitelist = {
                intent: set(values) for intent, values in iteritems(whitelist)}
        # pylint:enable=protected-access
        return parser


def _get_entity_scopes(dataset):
    intent_entities = extract_intent_entities(dataset)
    intent_groups = []
    entity_scopes = []
    for intent, entities in sorted(iteritems(intent_entities)):
        scope = {
            "builtin": list(
                {ent for ent in entities if is_builtin_entity(ent)}),
            "custom": list(
                {ent for ent in entities if not is_builtin_entity(ent)})
        }
        if scope in entity_scopes:
            group_idx = entity_scopes.index(scope)
            intent_groups[group_idx].append(intent)
        else:
            entity_scopes.append(scope)
            intent_groups.append([intent])
    return [
        {
            "intent_group": intent_group,
            "entity_scope": entity_scope
        } for intent_group, entity_scope in zip(intent_groups, entity_scopes)
    ]


def _get_entity_placeholders(dataset, language):
    return {
        e: _get_entity_name_placeholder(e, language) for e in dataset[ENTITIES]
    }


def _get_entity_name_placeholder(entity_label, language):
    return "%%%s%%" % "".join(tokenize_light(entity_label, language)).upper()


def _convert_dict_keys_to_int(dct):
    if isinstance(dct, dict):
        return {int(k): v for k, v in iteritems(dct)}
    return dct


def _get_entities_combinations(entities):
    yield ()
    for nb_entities in reversed(range(1, len(entities) + 1)):
        for combination in combinations(entities, nb_entities):
            yield combination



from snips_nlu.__about__ import __model_version__


class SnipsNLUError(Exception):
    """Base class for exceptions raised in the snips-nlu library"""


class IncompatibleModelError(Exception):
    """Raised when trying to load an incompatible NLU engine

    This happens when the engine data was persisted with a previous version of
    the library which is not compatible with the one used to load the model.
    """

    def __init__(self, persisted_version):
        super(IncompatibleModelError, self).__init__(
            "Incompatible data model: persisted model=%s, python lib model=%s"
            % (persisted_version, __model_version__))


class InvalidInputError(Exception):
    """Raised when an incorrect input is passed to one of the APIs"""


class NotTrained(SnipsNLUError):
    """Raised when a processing unit is used while not fitted"""


class IntentNotFoundError(SnipsNLUError):
    """Raised when an intent is used although it was not part of the
    training data"""

    def __init__(self, intent):
        super(IntentNotFoundError, self).__init__("Unknown intent '%s'"
                                                  % intent)


class DatasetFormatError(SnipsNLUError):
    """Raised when attempting to create a Snips NLU dataset using a wrong
    format"""


class _EmptyDatasetUtterancesError(SnipsNLUError):
    """Raised when attempting to train a processing unit on a dataset having
     only empty utterances"""


class EntityFormatError(DatasetFormatError):
    """Raised when attempting to create a Snips NLU entity using a wrong
    format"""


class IntentFormatError(DatasetFormatError):
    """Raised when attempting to create a Snips NLU intent using a wrong
    format"""


class AlreadyRegisteredError(SnipsNLUError):
    """Raised when attempting to register a subclass which is already
    registered"""

    def __init__(self, name, new_class, existing_class):
        msg = "Cannot register %s for %s as it has already been used to " \
              "register %s" \
              % (name, new_class.__name__, existing_class.__name__)
        super(AlreadyRegisteredError, self).__init__(msg)


class NotRegisteredError(SnipsNLUError):
    """Raised when trying to use a subclass which was not registered"""

    def __init__(self, registrable_cls, name=None, registered_cls=None):
        if name is not None:
            msg = "'%s' has not been registered for type %s. " \
                  % (name, registrable_cls)
        else:
            msg = "subclass %s has not been registered for type %s. " \
                  % (registered_cls, registrable_cls)
        msg += "Use @BaseClass.register('my_component') to register a subclass"
        super(NotRegisteredError, self).__init__(msg)


class PersistingError(SnipsNLUError):
    """Raised when trying to persist a processing unit to a path which already
    exists"""

    def __init__(self, path):
        super(PersistingError, self).__init__("Path already exists: %s"
                                              % str(path))


class LoadingError(SnipsNLUError):
    """Raised when trying to load a processing unit while some files are
    missing"""



from __future__ import division, unicode_literals

import json
from builtins import str, zip
from copy import deepcopy
from pathlib import Path

from future.utils import iteritems

from snips_nlu.common.utils import (
    json_string, fitted_required, replace_entities_with_placeholders,
    check_persisted_path)
from snips_nlu.constants import (
    DATA, ENTITY, ENTITY_KIND, LANGUAGE, NGRAM, TEXT, ENTITIES)
from snips_nlu.dataset import get_text_from_chunks, validate_and_format_dataset
from snips_nlu.entity_parser.builtin_entity_parser import (
    is_builtin_entity)
from snips_nlu.exceptions import (_EmptyDatasetUtterancesError, LoadingError)
from snips_nlu.languages import get_default_sep
from snips_nlu.pipeline.configs import FeaturizerConfig
from snips_nlu.pipeline.configs.intent_classifier import (
    CooccurrenceVectorizerConfig, TfidfVectorizerConfig)
from snips_nlu.pipeline.processing_unit import ProcessingUnit
from snips_nlu.preprocessing import stem, tokenize_light
from snips_nlu.resources import get_stop_words, get_word_cluster
from snips_nlu.slot_filler.features_utils import get_all_ngrams


@ProcessingUnit.register("featurizer")
class Featurizer(ProcessingUnit):
    """Feature extractor for text classification relying on ngrams tfidf and
    optionally word cooccurrences features"""

    config_type = FeaturizerConfig

    def __init__(self, config=None, **shared):
        super(Featurizer, self).__init__(config, **shared)
        self.language = None
        self.tfidf_vectorizer = None
        self.cooccurrence_vectorizer = None

    @property
    def fitted(self):
        if not self.tfidf_vectorizer or not self.tfidf_vectorizer.vocabulary:
            return False
        return True

    @property
    def feature_index_to_feature_name(self):
        """Maps the feature index of the feature matrix to printable features
        names. Mainly useful for debug.

        Returns:
            dict: a dict mapping feature indices to printable features names
        """
        if not self.fitted:
            return dict()

        index = {
            i: "ngram:%s" % ng
            for ng, i in iteritems(self.tfidf_vectorizer.vocabulary)
        }
        num_ng = len(index)
        if self.cooccurrence_vectorizer is not None:
            for word_pair, j in iteritems(
                    self.cooccurrence_vectorizer.word_pairs):
                index[j + num_ng] = "pair:%s+%s" % (word_pair[0], word_pair[1])
        return index

    def fit(self, dataset, utterances, classes, none_class):
        self.fit_transform(dataset, utterances, classes, none_class)
        return self

    def fit_transform(self, dataset, utterances, classes, none_class):
        import scipy.sparse as sp

        dataset = validate_and_format_dataset(dataset)
        self.language = dataset[LANGUAGE]

        utterances_texts = (get_text_from_chunks(u[DATA]) for u in utterances)
        if not any(tokenize_light(q, self.language) for q in utterances_texts):
            raise _EmptyDatasetUtterancesError(
                "Tokenized utterances are empty")

        x_tfidf = self._fit_transform_tfidf_vectorizer(
            utterances, classes, dataset)
        x = x_tfidf
        if self.config.added_cooccurrence_feature_ratio:
            self._fit_cooccurrence_vectorizer(
                utterances, classes, none_class, dataset)
            x_cooccurrence = self.cooccurrence_vectorizer.transform(utterances)
            x = sp.hstack((x_tfidf, x_cooccurrence))

        return x

    def transform(self, utterances):
        import scipy.sparse as sp

        x = self.tfidf_vectorizer.transform(utterances)
        if self.cooccurrence_vectorizer:
            x_cooccurrence = self.cooccurrence_vectorizer.transform(utterances)
            x = sp.hstack((x, x_cooccurrence))
        return x

    def _fit_transform_tfidf_vectorizer(self, x, y, dataset):
        from sklearn.feature_selection import chi2

        self.tfidf_vectorizer = TfidfVectorizer(
            config=self.config.tfidf_vectorizer_config,
            builtin_entity_parser=self.builtin_entity_parser,
            custom_entity_parser=self.custom_entity_parser,
            resources=self.resources,
            random_state=self.random_state,
        )
        x_tfidf = self.tfidf_vectorizer.fit_transform(x, dataset)

        if not self.tfidf_vectorizer.vocabulary:
            raise _EmptyDatasetUtterancesError(
                "Dataset is empty or with empty utterances")
        _, tfidf_pval = chi2(x_tfidf, y)
        best_tfidf_features = set(i for i, v in enumerate(tfidf_pval)
                                  if v < self.config.pvalue_threshold)
        if not best_tfidf_features:
            best_tfidf_features = set(
                idx for idx, val in enumerate(tfidf_pval) if
                val == tfidf_pval.min())

        best_ngrams = [ng for ng, i in
                       iteritems(self.tfidf_vectorizer.vocabulary)
                       if i in best_tfidf_features]
        self.tfidf_vectorizer.limit_vocabulary(best_ngrams)
        # We can't return x_tfidf[:best_tfidf_features] because of the
        # normalization in the transform of the tfidf_vectorizer
        # this would lead to inconsistent result between: fit_transform(x, y)
        # and fit(x, y).transform(x)
        return self.tfidf_vectorizer.transform(x)

    def _fit_cooccurrence_vectorizer(self, x, classes, none_class, dataset):
        import numpy as np
        from sklearn.feature_selection import chi2

        non_null_x = (d for d, c in zip(x, classes) if c != none_class)
        self.cooccurrence_vectorizer = CooccurrenceVectorizer(
            config=self.config.cooccurrence_vectorizer_config,
            builtin_entity_parser=self.builtin_entity_parser,
            custom_entity_parser=self.custom_entity_parser,
            resources=self.resources,
            random_state=self.random_state,
        )
        x_cooccurrence = self.cooccurrence_vectorizer.fit(
            non_null_x, dataset).transform(x)
        if not self.cooccurrence_vectorizer.word_pairs:
            return self
        _, pval = chi2(x_cooccurrence, classes)

        top_k = int(self.config.added_cooccurrence_feature_ratio * len(
            self.tfidf_vectorizer.idf_diag))

        # No selection if k is greater or equal than the number of word pairs
        if top_k >= len(self.cooccurrence_vectorizer.word_pairs):
            return self

        top_k_cooccurrence_ix = np.argpartition(
            pval, top_k - 1, axis=None)[:top_k]
        top_k_cooccurrence_ix = set(top_k_cooccurrence_ix)
        top_word_pairs = [
            pair for pair, i in iteritems(
                self.cooccurrence_vectorizer.word_pairs)
            if i in top_k_cooccurrence_ix
        ]

        self.cooccurrence_vectorizer.limit_word_pairs(top_word_pairs)
        return self

    @check_persisted_path
    def persist(self, path):
        path.mkdir()

        # Persist the vectorizers
        tfidf_vectorizer = None
        if self.tfidf_vectorizer:
            tfidf_vectorizer = self.tfidf_vectorizer.unit_name
            tfidf_vectorizer_path = path / tfidf_vectorizer
            self.tfidf_vectorizer.persist(tfidf_vectorizer_path)

        cooccurrence_vectorizer = None
        if self.cooccurrence_vectorizer:
            cooccurrence_vectorizer = self.cooccurrence_vectorizer.unit_name
            cooccurrence_vectorizer_path = path / cooccurrence_vectorizer
            self.cooccurrence_vectorizer.persist(cooccurrence_vectorizer_path)

        # Persist main object
        self_as_dict = {
            "language_code": self.language,
            "tfidf_vectorizer": tfidf_vectorizer,
            "cooccurrence_vectorizer": cooccurrence_vectorizer,
            "config": self.config.to_dict()
        }

        featurizer_path = path / "featurizer.json"
        with featurizer_path.open("w", encoding="utf-8") as f:
            f.write(json_string(self_as_dict))

        # Persist metadata
        self.persist_metadata(path)

    @classmethod
    def from_path(cls, path, **shared):
        path = Path(path)

        model_path = path / "featurizer.json"
        if not model_path.exists():
            raise LoadingError("Missing featurizer model file: %s"
                               % model_path.name)
        with model_path.open("r", encoding="utf-8") as f:
            featurizer_dict = json.load(f)

        featurizer_config = featurizer_dict["config"]
        featurizer = cls(featurizer_config, **shared)

        featurizer.language = featurizer_dict["language_code"]

        tfidf_vectorizer = featurizer_dict["tfidf_vectorizer"]
        if tfidf_vectorizer:
            vectorizer_path = path / featurizer_dict["tfidf_vectorizer"]
            tfidf_vectorizer = TfidfVectorizer.from_path(
                vectorizer_path, **shared)
        featurizer.tfidf_vectorizer = tfidf_vectorizer

        cooccurrence_vectorizer = featurizer_dict["cooccurrence_vectorizer"]
        if cooccurrence_vectorizer:
            vectorizer_path = path / featurizer_dict["cooccurrence_vectorizer"]
            cooccurrence_vectorizer = CooccurrenceVectorizer.from_path(
                vectorizer_path, **shared)
        featurizer.cooccurrence_vectorizer = cooccurrence_vectorizer

        return featurizer


@ProcessingUnit.register("tfidf_vectorizer")
class TfidfVectorizer(ProcessingUnit):
    """Wrapper of the scikit-learn TfidfVectorizer"""

    config_type = TfidfVectorizerConfig

    def __init__(self, config=None, **shared):
        super(TfidfVectorizer, self).__init__(config, **shared)
        self._tfidf_vectorizer = None
        self._language = None
        self.builtin_entity_scope = None

    def fit(self, x, dataset):
        """Fits the idf of the vectorizer on the given utterances after
        enriching them with builtin entities matches, custom entities matches
        and the potential word clusters matches

        Args:
            x (list of dict): list of utterances
            dataset (dict): dataset from which x was extracted (needed to
                extract the language and the builtin entity scope)

        Returns:
            :class:`.TfidfVectorizer`: The fitted vectorizer
        """
        self.load_resources_if_needed(dataset[LANGUAGE])
        self.fit_builtin_entity_parser_if_needed(dataset)
        self.fit_custom_entity_parser_if_needed(dataset)

        self._language = dataset[LANGUAGE]
        self._init_vectorizer(self._language)
        self.builtin_entity_scope = set(
            e for e in dataset[ENTITIES] if is_builtin_entity(e))
        preprocessed_data = self._preprocess(x)
        utterances = [
            self._enrich_utterance(u, builtin_ents, custom_ents, w_clusters)
            for u, builtin_ents, custom_ents, w_clusters
            in zip(*preprocessed_data)
        ]
        return self._tfidf_vectorizer.fit(utterances)

    def fit_transform(self, x, dataset):
        """Fits the idf of the vectorizer on the given utterances after
        enriching them with builtin entities matches, custom entities matches
        and the potential word clusters matches.
        Returns the featurized utterances.

        Args:
            x (list of dict): list of utterances
            dataset (dict): dataset from which x was extracted (needed to
                extract the language and the builtin entity scope)

        Returns:
            :class:`.scipy.sparse.csr_matrix`: A sparse matrix X of shape
            (len(x), len(self.vocabulary)) where X[i, j] contains tfdif of
            the ngram of index j of the vocabulary in the utterance i
        """
        self.load_resources_if_needed(dataset[LANGUAGE])
        self.fit_builtin_entity_parser_if_needed(dataset)
        self.fit_custom_entity_parser_if_needed(dataset)

        self._language = dataset[LANGUAGE]
        self._init_vectorizer(self._language)
        self.builtin_entity_scope = set(
            e for e in dataset[ENTITIES] if is_builtin_entity(e))
        preprocessed_data = self._preprocess(x)
        utterances = [
            self._enrich_utterance(u, builtin_ents, custom_ents, w_clusters)
            for u, builtin_ents, custom_ents, w_clusters
            in zip(*preprocessed_data)
        ]
        return self._tfidf_vectorizer.fit_transform(utterances)

    @property
    def fitted(self):
        return self._tfidf_vectorizer is not None and hasattr(
            self._tfidf_vectorizer, "vocabulary_")

    @fitted_required
    def transform(self, x):
        """Featurizes the given utterances after enriching them with builtin
        entities matches, custom entities matches and the potential word
        clusters matches

        Args:
            x (list of dict): list of utterances

        Returns:
            :class:`.scipy.sparse.csr_matrix`: A sparse matrix X of shape
            (len(x), len(self.vocabulary)) where X[i, j] contains tfdif of
            the ngram of index j of the vocabulary in the utterance i

        Raises:
            NotTrained: when the vectorizer is not fitted:
        """
        utterances = [self._enrich_utterance(*data)
                      for data in zip(*self._preprocess(x))]
        return self._tfidf_vectorizer.transform(utterances)

    def _preprocess(self, utterances):
        normalized_utterances = deepcopy(utterances)
        for u in normalized_utterances:
            nb_chunks = len(u[DATA])
            for i, chunk in enumerate(u[DATA]):
                chunk[TEXT] = _normalize_stem(
                    chunk[TEXT], self.language, self.resources,
                    self.config.use_stemming)
                if i < nb_chunks - 1:
                    chunk[TEXT] += " "

        # Extract builtin entities on unormalized utterances
        builtin_ents = [
            self.builtin_entity_parser.parse(
                get_text_from_chunks(u[DATA]),
                self.builtin_entity_scope, use_cache=True)
            for u in utterances
        ]
        # Extract builtin entities on normalized utterances
        custom_ents = [
            self.custom_entity_parser.parse(
                get_text_from_chunks(u[DATA]), use_cache=True)
            for u in normalized_utterances
        ]
        if self.config.word_clusters_name:
            # Extract world clusters on unormalized utterances
            original_utterances_text = [get_text_from_chunks(u[DATA])
                                        for u in utterances]
            w_clusters = [
                _get_word_cluster_features(
                    tokenize_light(u.lower(), self.language),
                    self.config.word_clusters_name,
                    self.resources)
                for u in original_utterances_text
            ]
        else:
            w_clusters = [None for _ in normalized_utterances]

        return normalized_utterances, builtin_ents, custom_ents, w_clusters

    def _enrich_utterance(self, utterance, builtin_entities, custom_entities,
                          word_clusters):
        custom_entities_features = [
            _entity_name_to_feature(e[ENTITY_KIND], self.language)
            for e in custom_entities]

        builtin_entities_features = [
            _builtin_entity_to_feature(ent[ENTITY_KIND], self.language)
            for ent in builtin_entities
        ]

        # We remove values of builtin slots from the utterance to avoid
        # learning specific samples such as '42' or 'tomorrow'
        filtered_tokens = [
            chunk[TEXT] for chunk in utterance[DATA]
            if ENTITY not in chunk or not is_builtin_entity(chunk[ENTITY])
        ]

        features = get_default_sep(self.language).join(filtered_tokens)

        if builtin_entities_features:
            features += " " + " ".join(sorted(builtin_entities_features))
        if custom_entities_features:
            features += " " + " ".join(sorted(custom_entities_features))
        if word_clusters:
            features += " " + " ".join(sorted(word_clusters))

        return features

    @property
    def language(self):
        # Create this getter to prevent the language from being set elsewhere
        # than in the fit
        return self._language

    @property
    def vocabulary(self):
        if self._tfidf_vectorizer and hasattr(
                self._tfidf_vectorizer, "vocabulary_"):
            return self._tfidf_vectorizer.vocabulary_
        return None

    @fitted_required
    def limit_vocabulary(self, ngrams):
        """Restrict the vectorizer vocabulary to the given ngrams

        Args:
            ngrams (iterable of str or tuples of str): ngrams to keep

        Returns:
            :class:`.TfidfVectorizer`: The vectorizer with limited vocabulary
        """
        import scipy.sparse as sp

        ngrams = set(ngrams)
        vocab = self.vocabulary
        existing_ngrams = set(vocab)
        extra_values = ngrams - existing_ngrams

        if extra_values:
            raise ValueError("Invalid ngrams %s, expected values in word_pairs"
                             % sorted(extra_values))

        new_ngrams, new_index = zip(*sorted((ng, vocab[ng]) for ng in ngrams))

        self._tfidf_vectorizer.vocabulary_ = {
            ng: new_i for new_i, ng in enumerate(new_ngrams)
        }
        # pylint: disable=protected-access
        # The new_idf_data is valid because the previous _idf_diag was indexed
        # with sorted ngrams and new_index is also indexed with sorted ngrams
        new_idf_data = self._tfidf_vectorizer._tfidf._idf_diag.data[
            list(new_index)]
        self._tfidf_vectorizer._tfidf._idf_diag = sp.spdiags(
            new_idf_data, diags=0, m=len(new_index), n=len(new_index),
            format="csr")
        # pylint: enable=protected-access
        return self

    @property
    def idf_diag(self):
        if self._tfidf_vectorizer and hasattr(
                self._tfidf_vectorizer, "vocabulary_"):
            return self._tfidf_vectorizer.idf_
        return None

    def _init_vectorizer(self, language):
        from sklearn.feature_extraction.text import (
            TfidfVectorizer as SklearnTfidfVectorizer)

        self._tfidf_vectorizer = SklearnTfidfVectorizer(
            tokenizer=lambda x: tokenize_light(x, language))
        return self

    @check_persisted_path
    def persist(self, path):
        path.mkdir()

        vectorizer_ = None
        if self._tfidf_vectorizer is not None:
            vocab = {k: int(v) for k, v in iteritems(self.vocabulary)}
            idf_diag = self.idf_diag.tolist()
            vectorizer_ = {
                "vocab": vocab,
                "idf_diag": idf_diag
            }

        builtin_entity_scope = None
        if self.builtin_entity_scope is not None:
            builtin_entity_scope = list(self.builtin_entity_scope)

        self_as_dict = {
            "vectorizer": vectorizer_,
            "language_code": self.language,
            "builtin_entity_scope": builtin_entity_scope,
            "config": self.config.to_dict(),
        }

        vectorizer_path = path / "vectorizer.json"
        with vectorizer_path.open("w", encoding="utf-8") as f:
            f.write(json_string(self_as_dict))
        self.persist_metadata(path)

    @classmethod
    # pylint: disable=W0212
    def from_path(cls, path, **shared):
        import numpy as np
        import scipy.sparse as sp
        from sklearn.feature_extraction.text import (
            TfidfTransformer, TfidfVectorizer as SklearnTfidfVectorizer)

        path = Path(path)

        model_path = path / "vectorizer.json"
        if not model_path.exists():
            raise LoadingError("Missing vectorizer model file: %s"
                               % model_path.name)
        with model_path.open("r", encoding="utf-8") as f:
            vectorizer_dict = json.load(f)

        vectorizer = cls(vectorizer_dict["config"], **shared)
        vectorizer._language = vectorizer_dict["language_code"]

        builtin_entity_scope = vectorizer_dict["builtin_entity_scope"]
        if builtin_entity_scope is not None:
            builtin_entity_scope = set(builtin_entity_scope)
        vectorizer.builtin_entity_scope = builtin_entity_scope

        vectorizer_ = vectorizer_dict["vectorizer"]
        if vectorizer_:
            vocab = vectorizer_["vocab"]
            idf_diag_data = vectorizer_["idf_diag"]
            idf_diag_data = np.array(idf_diag_data)

            idf_diag_shape = (len(idf_diag_data), len(idf_diag_data))
            row = list(range(idf_diag_shape[0]))
            col = list(range(idf_diag_shape[0]))
            idf_diag = sp.csr_matrix(
                (idf_diag_data, (row, col)), shape=idf_diag_shape)

            tfidf_transformer = TfidfTransformer()
            tfidf_transformer._idf_diag = idf_diag

            vectorizer_ = SklearnTfidfVectorizer(
                tokenizer=lambda x: tokenize_light(x, vectorizer._language))
            vectorizer_.vocabulary_ = vocab

            vectorizer_._tfidf = tfidf_transformer

        vectorizer._tfidf_vectorizer = vectorizer_
        return vectorizer


@ProcessingUnit.register("cooccurrence_vectorizer")
class CooccurrenceVectorizer(ProcessingUnit):
    """Featurizer that takes utterances and extracts ordered word cooccurrence
     features matrix from them"""

    config_type = CooccurrenceVectorizerConfig

    def __init__(self, config=None, **shared):
        super(CooccurrenceVectorizer, self).__init__(config, **shared)
        self._word_pairs = None
        self._language = None
        self.builtin_entity_scope = None

    @property
    def language(self):
        # Create this getter to prevent the language from being set elsewhere
        # than in the fit
        return self._language

    @property
    def word_pairs(self):
        return self._word_pairs

    def fit(self, x, dataset):
        """Fits the CooccurrenceVectorizer

        Given a list of utterances the CooccurrenceVectorizer will extract word
        pairs appearing in the same utterance. The order in which the words
        appear is kept. Additionally, if self.config.window_size is not None
        then the vectorizer will only look in a context window of
        self.config.window_size after each word.

        Args:
            x (iterable): list of utterances
            dataset (dict): dataset from which x was extracted (needed to
                extract the language and the builtin entity scope)

        Returns:
            :class:`.CooccurrenceVectorizer`: The fitted vectorizer
        """
        self.load_resources_if_needed(dataset[LANGUAGE])
        self.fit_builtin_entity_parser_if_needed(dataset)
        self.fit_custom_entity_parser_if_needed(dataset)

        self._language = dataset[LANGUAGE]
        self.builtin_entity_scope = set(
            e for e in dataset[ENTITIES] if is_builtin_entity(e))

        preprocessed = self._preprocess(list(x))
        utterances = [
            self._enrich_utterance(utterance, builtin_ents, custom_ent)
            for utterance, builtin_ents, custom_ent in zip(*preprocessed)]
        word_pairs = set(
            p for u in utterances for p in self._extract_word_pairs(u))
        self._word_pairs = {
            pair: i for i, pair in enumerate(sorted(word_pairs))
        }
        return self

    @property
    def fitted(self):
        """Whether or not the vectorizer is fitted"""
        return self.word_pairs is not None

    def fit_transform(self, x, dataset):
        """Fits the vectorizer and returns the feature matrix

        Args:
            x (iterable): iterable of 3-tuples of the form
                (tokenized_utterances, builtin_entities, custom_entities)
            dataset (dict): dataset from which x was extracted (needed to
                extract the language and the builtin entity scope)

        Returns:
            :class:`.scipy.sparse.csr_matrix`: A sparse matrix X of shape
            (len(x), len(self.word_pairs)) where
            X[i, j] = 1.0 if x[i][0] contains the words cooccurrence
            (w1, w2) and if self.word_pairs[(w1, w2)] = j
        """
        return self.fit(x, dataset).transform(x)

    def _enrich_utterance(self, x, builtin_ents, custom_ents):
        utterance = get_text_from_chunks(x[DATA])
        all_entities = builtin_ents + custom_ents
        placeholder_fn = self._placeholder_fn
        # Replace entities with placeholders
        enriched_utterance = replace_entities_with_placeholders(
            utterance, all_entities, placeholder_fn)[1]
        # Tokenize
        enriched_utterance = tokenize_light(enriched_utterance, self.language)
        # Remove the unknownword strings if needed
        if self.config.unknown_words_replacement_string:
            enriched_utterance = [
                t for t in enriched_utterance
                if t != self.config.unknown_words_replacement_string
            ]
        return enriched_utterance

    @fitted_required
    def transform(self, x):
        """Computes the cooccurrence feature matrix.

        Args:
            x (list of dict): list of utterances

        Returns:
            :class:`.scipy.sparse.csr_matrix`: A sparse matrix X of shape
            (len(x), len(self.word_pairs)) where X[i, j] = 1.0 if
            x[i][0] contains the words cooccurrence (w1, w2) and if
            self.word_pairs[(w1, w2)] = j

        Raises:
            NotTrained: when the vectorizer is not fitted
        """
        import numpy as np
        import scipy.sparse as sp

        preprocessed = self._preprocess(x)
        utterances = [
            self._enrich_utterance(utterance, builtin_ents, custom_ent)
            for utterance, builtin_ents, custom_ent in zip(*preprocessed)]

        x_coo = sp.dok_matrix((len(x), len(self.word_pairs)), dtype=np.int32)
        for i, u in enumerate(utterances):
            for p in self._extract_word_pairs(u):
                if p in self.word_pairs:
                    x_coo[i, self.word_pairs[p]] = 1

        return x_coo.tocsr()

    def _preprocess(self, x):
        # Extract all entities on unnormalized data
        builtin_ents = [
            self.builtin_entity_parser.parse(
                get_text_from_chunks(u[DATA]),
                self.builtin_entity_scope,
                use_cache=True
            ) for u in x
        ]
        custom_ents = [
            self.custom_entity_parser.parse(
                get_text_from_chunks(u[DATA]), use_cache=True)
            for u in x
        ]
        return x, builtin_ents, custom_ents

    def _extract_word_pairs(self, utterance):
        if self.config.filter_stop_words:
            stop_words = get_stop_words(self.resources)
            utterance = [t for t in utterance if t not in stop_words]
        pairs = set()
        for j, w1 in enumerate(utterance):
            max_index = None
            if self.config.window_size is not None:
                max_index = j + self.config.window_size + 1
            for w2 in utterance[j + 1:max_index]:
                key = (w1, w2)
                if not self.config.keep_order:
                    key = tuple(sorted(key))
                pairs.add(key)
        return pairs

    @fitted_required
    def limit_word_pairs(self, word_pairs):
        """Restrict the vectorizer word pairs to the given word pairs

        Args:
            word_pairs (iterable of 2-tuples (str, str)): word_pairs to keep

        Returns:
            :class:`.CooccurrenceVectorizer`: The vectorizer with limited
            word pairs
        """
        word_pairs = set(word_pairs)
        existing_pairs = set(self.word_pairs)
        extra_values = word_pairs - existing_pairs

        if extra_values:
            raise ValueError(
                "Invalid word pairs %s, expected values in word_pairs"
                % sorted(extra_values))

        self._word_pairs = {
            ng: new_i for new_i, ng in enumerate(sorted(word_pairs))
        }
        return self

    def _placeholder_fn(self, entity_name):
        return "".join(
            tokenize_light(str(entity_name), str(self.language))).upper()

    @check_persisted_path
    def persist(self, path):
        path.mkdir()

        builtin_entity_scope = None
        if self.builtin_entity_scope is not None:
            builtin_entity_scope = list(self.builtin_entity_scope)

        self_as_dict = {
            "language_code": self.language,
            "word_pairs": {
                i: list(p) for p, i in iteritems(self.word_pairs)
            },
            "builtin_entity_scope": builtin_entity_scope,
            "config": self.config.to_dict()
        }
        vectorizer_json = json_string(self_as_dict)
        vectorizer_path = path / "vectorizer.json"
        with vectorizer_path.open(mode="w", encoding="utf8") as f:
            f.write(vectorizer_json)
        self.persist_metadata(path)

    @classmethod
    # pylint: disable=protected-access
    def from_path(cls, path, **shared):
        path = Path(path)
        model_path = path / "vectorizer.json"
        if not model_path.exists():
            raise LoadingError("Missing vectorizer model file: %s"
                               % model_path.name)

        with model_path.open(encoding="utf8") as f:
            vectorizer_dict = json.load(f)
        config = vectorizer_dict.pop("config")

        self = cls(config, **shared)
        self._language = vectorizer_dict["language_code"]
        self._word_pairs = None

        builtin_entity_scope = vectorizer_dict["builtin_entity_scope"]
        if builtin_entity_scope is not None:
            builtin_entity_scope = set(builtin_entity_scope)
        self.builtin_entity_scope = builtin_entity_scope

        if vectorizer_dict["word_pairs"]:
            self._word_pairs = {
                tuple(p): int(i)
                for i, p in iteritems(vectorizer_dict["word_pairs"])
            }
        return self


def _entity_name_to_feature(entity_name, language):
    return "entityfeature%s" % "".join(tokenize_light(
        entity_name.lower(), language))


def _builtin_entity_to_feature(builtin_entity_label, language):
    return "builtinentityfeature%s" % "".join(tokenize_light(
        builtin_entity_label.lower(), language))


def _normalize_stem(text, language, resources, use_stemming):
    from snips_nlu_utils import normalize

    if use_stemming:
        return stem(text, language, resources)
    return normalize(text)


def _get_word_cluster_features(query_tokens, clusters_name, resources):
    if not clusters_name:
        return []
    ngrams = get_all_ngrams(query_tokens)
    cluster_features = []
    for ngram in ngrams:
        cluster = get_word_cluster(resources, clusters_name).get(
            ngram[NGRAM].lower(), None)
        if cluster is not None:
            cluster_features.append(cluster)
    return cluster_features



from __future__ import print_function, unicode_literals


def add_download_entity_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "download-entity", formatter_class=formatter_class,
        help="Download resources for a builtin gazetteer entity")
    subparser.add_argument(
        "entity_name", type=str,
        help="Name of the builtin entity to download, e.g. snips/musicArtist")
    subparser.add_argument("language", type=str,
                           help="Language of the builtin entity")
    subparser.add_argument(
        "extra_pip_args", nargs="*", type=str,
        help="Additional arguments to be passed to `pip install` when "
             "installing the builtin entity package")
    subparser.set_defaults(func=_download_builtin_entity)
    return subparser


def _download_builtin_entity(args_namespace):
    return download_builtin_entity(
        args_namespace.entity_name, args_namespace.language,
        *args_namespace.extra_pip_args)


# inspired from
# https://github.com/explosion/spaCy/blob/master/spacy/cli/download.py

def download_builtin_entity(entity_name, language, *pip_args):
    """Download compatible language or gazetteer entity resources"""
    from snips_nlu import __about__
    from snips_nlu.cli.download import download_from_resource_name
    from snips_nlu.cli.utils import (
        check_resources_alias, get_compatibility, get_json)

    download_from_resource_name(language, pip_args, verbose=False)

    shortcuts = get_json(__about__.__shortcuts__, "Resource shortcuts")
    check_resources_alias(entity_name, shortcuts)

    compatibility = get_compatibility()
    resource_name_lower = entity_name.lower()
    long_resource_name = shortcuts.get(resource_name_lower,
                                       resource_name_lower)

    _download_and_link_entity(
        long_resource_name, entity_name, language, compatibility,
        pip_args)


def add_download_language_entities_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "download-language-entities", formatter_class=formatter_class,
        help="Download resources for all builtin gazetteer entities in a "
             "given language")
    subparser.add_argument("language", type=str,
                           help="Language of the builtin entities")
    subparser.add_argument(
        "extra_pip_args", nargs="*", type=str,
        help="Additional arguments to be passed to `pip install` when "
             "installing the builtin entities packages")
    subparser.set_defaults(func=_download_language_builtin_entities)
    return subparser


def _download_language_builtin_entities(args_namespace):
    return download_language_builtin_entities(
        args_namespace.language, *args_namespace.extra_pip_args)


def download_language_builtin_entities(language, *pip_args):
    """Download all gazetteer entity resources for a given language as well as
    basic language resources for this language"""
    from builtins import str
    from snips_nlu_parsers import get_supported_gazetteer_entities
    from snips_nlu import __about__
    from snips_nlu.cli.download import download_from_resource_name
    from snips_nlu.cli.utils import (
        check_resources_alias, get_compatibility, get_json)

    download_from_resource_name(language, pip_args, verbose=False)

    shortcuts = get_json(__about__.__shortcuts__, "Resource shortcuts")
    for entity_name in get_supported_gazetteer_entities(str(language)):
        check_resources_alias(entity_name, shortcuts)

        compatibility = get_compatibility()
        resource_name_lower = entity_name.lower()
        long_resource_name = shortcuts.get(resource_name_lower,
                                           resource_name_lower)

        _download_and_link_entity(
            long_resource_name, entity_name, language, compatibility,
            pip_args)


def _download_and_link_entity(long_resource_name, entity_name, language,
                              compatibility, pip_args):
    import sys
    from builtins import str
    from snips_nlu_parsers import get_builtin_entity_shortname
    from snips_nlu.cli.link import link_resources
    from snips_nlu.cli.utils import (
        PrettyPrintLevel, get_json, get_resources_version,
        install_remote_package, pretty_print)
    from snips_nlu.common.utils import get_package_path

    full_resource_name = long_resource_name + "_" + language
    version = get_resources_version(full_resource_name, entity_name,
                                    compatibility)
    entity_alias = get_builtin_entity_shortname(str(entity_name)).lower()
    entity_base_url = _get_entity_base_url(language, entity_alias, version)
    latest = get_json(entity_base_url + "/latest",
                      "Latest entity resources version")
    latest_url = "{b}/{n}#egg={r}=={v}".format(
        b=entity_base_url, n=latest["filename"], r=full_resource_name,
        v=latest["version"])
    exit_code = install_remote_package(latest_url, pip_args)
    if exit_code != 0:
        sys.exit(exit_code)
    try:
        # Get package path here because link uses
        # pip.get_installed_distributions() to check if the resource is a
        # package, which fails if the resource was just installed via
        # subprocess
        package_path = get_package_path(full_resource_name)
        link_alias = entity_alias + "_" + language
        link_path, resources_dir = link_resources(
            full_resource_name, link_alias, force=True,
            resources_path=package_path)
        pretty_print("%s --> %s" % (str(resources_dir), str(link_path)),
                     "You can now use the '%s' builtin entity" % entity_name,
                     title="Linking successful",
                     level=PrettyPrintLevel.SUCCESS)
    except:  # pylint:disable=bare-except
        pretty_print(
            "Creating a shortcut link for '%s' didn't work." % entity_name,
            title="The builtin entity resources were successfully downloaded, "
                  "however linking failed.",
            level=PrettyPrintLevel.ERROR)


def _get_entity_base_url(language, entity_alias, version):
    from snips_nlu import __about__

    if not version.startswith("v"):
        version = "v" + version
    return "/".join(
        [__about__.__entities_download_url__, language, entity_alias, version])



from __future__ import unicode_literals

import json
import logging
from builtins import range, str, zip
from pathlib import Path

from snips_nlu.common.log_utils import DifferedLoggingMessage, log_elapsed_time
from snips_nlu.common.utils import (
    check_persisted_path, fitted_required, json_string)
from snips_nlu.constants import LANGUAGE, RES_PROBA
from snips_nlu.dataset import validate_and_format_dataset
from snips_nlu.exceptions import LoadingError, _EmptyDatasetUtterancesError
from snips_nlu.intent_classifier.featurizer import Featurizer
from snips_nlu.intent_classifier.intent_classifier import IntentClassifier
from snips_nlu.intent_classifier.log_reg_classifier_utils import (
    build_training_data, get_regularization_factor, text_to_utterance)
from snips_nlu.pipeline.configs import LogRegIntentClassifierConfig
from snips_nlu.result import intent_classification_result

logger = logging.getLogger(__name__)

# We set tol to 1e-3 to silence the following warning with Python 2 (
# scikit-learn 0.20):
#
# FutureWarning: max_iter and tol parameters have been added in SGDClassifier
# in 0.19. If max_iter is set but tol is left unset, the default value for tol
# in 0.19 and 0.20 will be None (which is equivalent to -infinity, so it has no
# effect) but will change in 0.21 to 1e-3. Specify tol to silence this warning.

LOG_REG_ARGS = {
    "loss": "log",
    "penalty": "l2",
    "max_iter": 1000,
    "tol": 1e-3,
    "n_jobs": -1
}


@IntentClassifier.register("log_reg_intent_classifier")
class LogRegIntentClassifier(IntentClassifier):
    """Intent classifier which uses a Logistic Regression underneath"""

    config_type = LogRegIntentClassifierConfig

    def __init__(self, config=None, **shared):
        """The LogReg intent classifier can be configured by passing a
        :class:`.LogRegIntentClassifierConfig`"""
        super(LogRegIntentClassifier, self).__init__(config, **shared)
        self.classifier = None
        self.intent_list = None
        self.featurizer = None

    @property
    def fitted(self):
        """Whether or not the intent classifier has already been fitted"""
        return self.intent_list is not None

    @log_elapsed_time(logger, logging.INFO,
                      "LogRegIntentClassifier in {elapsed_time}")
    def fit(self, dataset):
        """Fits the intent classifier with a valid Snips dataset

        Returns:
            :class:`LogRegIntentClassifier`: The same instance, trained
        """
        from sklearn.linear_model import SGDClassifier
        from sklearn.utils import compute_class_weight

        logger.info("Fitting LogRegIntentClassifier...")
        dataset = validate_and_format_dataset(dataset)
        self.load_resources_if_needed(dataset[LANGUAGE])
        self.fit_builtin_entity_parser_if_needed(dataset)
        self.fit_custom_entity_parser_if_needed(dataset)
        language = dataset[LANGUAGE]

        data_augmentation_config = self.config.data_augmentation_config
        utterances, classes, intent_list = build_training_data(
            dataset, language, data_augmentation_config, self.resources,
            self.random_state)

        self.intent_list = intent_list
        if len(self.intent_list) <= 1:
            return self

        self.featurizer = Featurizer(
            config=self.config.featurizer_config,
            builtin_entity_parser=self.builtin_entity_parser,
            custom_entity_parser=self.custom_entity_parser,
            resources=self.resources,
            random_state=self.random_state,
        )
        self.featurizer.language = language

        none_class = max(classes)
        try:
            x = self.featurizer.fit_transform(
                dataset, utterances, classes, none_class)
        except _EmptyDatasetUtterancesError:
            logger.warning("No (non-empty) utterances found in dataset")
            self.featurizer = None
            return self

        alpha = get_regularization_factor(dataset)

        class_weights_arr = compute_class_weight(
            "balanced", range(none_class + 1), classes)
        # Re-weight the noise class
        class_weights_arr[-1] *= self.config.noise_reweight_factor
        class_weight = {idx: w for idx, w in enumerate(class_weights_arr)}

        self.classifier = SGDClassifier(
            random_state=self.random_state, alpha=alpha,
            class_weight=class_weight, **LOG_REG_ARGS)
        self.classifier.fit(x, classes)
        logger.debug("%s", DifferedLoggingMessage(self.log_best_features))
        return self

    @fitted_required
    def get_intent(self, text, intents_filter=None):
        """Performs intent classification on the provided *text*

        Args:
            text (str): Input
            intents_filter (str or list of str): When defined, it will find
                the most likely intent among the list, otherwise it will use
                the whole list of intents defined in the dataset

        Returns:
            dict or None: The most likely intent along with its probability or
            *None* if no intent was found

        Raises:
            :class:`snips_nlu.exceptions.NotTrained`: When the intent
                classifier is not fitted

        """
        return self._get_intents(text, intents_filter)[0]

    @fitted_required
    def get_intents(self, text):
        """Performs intent classification on the provided *text* and returns
        the list of intents ordered by decreasing probability

        The length of the returned list is exactly the number of intents in the
        dataset + 1 for the None intent

        Raises:
            :class:`snips_nlu.exceptions.NotTrained`: when the intent
                classifier is not fitted
        """
        return self._get_intents(text, intents_filter=None)

    def _get_intents(self, text, intents_filter):
        if isinstance(intents_filter, str):
            intents_filter = {intents_filter}
        elif isinstance(intents_filter, list):
            intents_filter = set(intents_filter)

        if not text or not self.intent_list or not self.featurizer:
            results = [intent_classification_result(None, 1.0)]
            results += [intent_classification_result(i, 0.0)
                        for i in self.intent_list if i is not None]
            return results

        if len(self.intent_list) == 1:
            return [intent_classification_result(self.intent_list[0], 1.0)]

        # pylint: disable=C0103
        X = self.featurizer.transform([text_to_utterance(text)])
        # pylint: enable=C0103
        proba_vec = self._predict_proba(X)
        logger.debug(
            "%s", DifferedLoggingMessage(self.log_activation_weights, text, X))
        results = [
            intent_classification_result(i, proba)
            for i, proba in zip(self.intent_list, proba_vec[0])
            if intents_filter is None or i is None or i in intents_filter]

        return sorted(results, key=lambda res: -res[RES_PROBA])

    def _predict_proba(self, X):  # pylint: disable=C0103
        import numpy as np

        self.classifier._check_proba()  # pylint: disable=W0212

        prob = self.classifier.decision_function(X)
        prob *= -1
        np.exp(prob, prob)
        prob += 1
        np.reciprocal(prob, prob)
        if prob.ndim == 1:
            return np.vstack([1 - prob, prob]).T
        return prob

    @check_persisted_path
    def persist(self, path):
        """Persists the object at the given path"""
        path.mkdir()

        featurizer = None
        if self.featurizer is not None:
            featurizer = "featurizer"
            featurizer_path = path / featurizer
            self.featurizer.persist(featurizer_path)

        coeffs = None
        intercept = None
        t_ = None
        if self.classifier is not None:
            coeffs = self.classifier.coef_.tolist()
            intercept = self.classifier.intercept_.tolist()
            t_ = self.classifier.t_

        self_as_dict = {
            "config": self.config.to_dict(),
            "coeffs": coeffs,
            "intercept": intercept,
            "t_": t_,
            "intent_list": self.intent_list,
            "featurizer": featurizer
        }

        classifier_json = json_string(self_as_dict)
        with (path / "intent_classifier.json").open(mode="w",
                                                    encoding="utf8") as f:
            f.write(classifier_json)
        self.persist_metadata(path)

    @classmethod
    def from_path(cls, path, **shared):
        """Loads a :class:`LogRegIntentClassifier` instance from a path

        The data at the given path must have been generated using
        :func:`~LogRegIntentClassifier.persist`
        """
        import numpy as np
        from sklearn.linear_model import SGDClassifier

        path = Path(path)
        model_path = path / "intent_classifier.json"
        if not model_path.exists():
            raise LoadingError("Missing intent classifier model file: %s"
                               % model_path.name)

        with model_path.open(encoding="utf8") as f:
            model_dict = json.load(f)

        # Create the classifier
        config = LogRegIntentClassifierConfig.from_dict(model_dict["config"])
        intent_classifier = cls(config=config, **shared)
        intent_classifier.intent_list = model_dict['intent_list']

        # Create the underlying SGD classifier
        sgd_classifier = None
        coeffs = model_dict['coeffs']
        intercept = model_dict['intercept']
        t_ = model_dict["t_"]
        if coeffs is not None and intercept is not None:
            sgd_classifier = SGDClassifier(**LOG_REG_ARGS)
            sgd_classifier.coef_ = np.array(coeffs)
            sgd_classifier.intercept_ = np.array(intercept)
            sgd_classifier.t_ = t_
        intent_classifier.classifier = sgd_classifier

        # Add the featurizer
        featurizer = model_dict['featurizer']
        if featurizer is not None:
            featurizer_path = path / featurizer
            intent_classifier.featurizer = Featurizer.from_path(
                featurizer_path, **shared)

        return intent_classifier

    def log_best_features(self, top_n=50):
        import numpy as np

        if not hasattr(self.featurizer, "feature_index_to_feature_name"):
            return None

        log = "Top {} features weights by intent:".format(top_n)
        index_to_feature = self.featurizer.feature_index_to_feature_name
        for intent_ix in range(self.classifier.coef_.shape[0]):
            intent_name = self.intent_list[intent_ix]
            log += "\n\n\nFor intent {}\n".format(intent_name)
            top_features_idx = np.argsort(
                np.absolute(self.classifier.coef_[intent_ix]))[::-1][:top_n]
            for feature_ix in top_features_idx:
                feature_name = index_to_feature[feature_ix]
                feature_weight = self.classifier.coef_[intent_ix, feature_ix]
                log += "\n{} -> {}".format(feature_name, feature_weight)
        return log

    def log_activation_weights(self, text, x, top_n=50):
        import numpy as np

        if not hasattr(self.featurizer, "feature_index_to_feature_name"):
            return None

        log = "\n\nTop {} feature activations for: \"{}\":\n".format(
            top_n, text)
        activations = np.multiply(
            self.classifier.coef_, np.asarray(x.todense()))
        abs_activation = np.absolute(activations).flatten().squeeze()

        if top_n > activations.size:
            top_n = activations.size

        top_n_activations_ix = np.argpartition(abs_activation, -top_n,
                                               axis=None)[-top_n:]
        top_n_activations_ix = np.unravel_index(
            top_n_activations_ix, activations.shape)

        index_to_feature = self.featurizer.feature_index_to_feature_name
        features_intent_and_activation = [
            (self.intent_list[i], index_to_feature[f], activations[i, f])
            for i, f in zip(*top_n_activations_ix)]

        features_intent_and_activation = sorted(
            features_intent_and_activation, key=lambda x: abs(x[2]),
            reverse=True)

        for intent, feature, activation in features_intent_and_activation:
            log += "\n\n\"{}\" -> ({}, {:.2f})".format(
                intent, feature, float(activation))
        log += "\n\n"
        return log



from __future__ import unicode_literals

import io
import json
import shutil
from abc import ABCMeta, abstractmethod, abstractproperty
from builtins import str, bytes
from pathlib import Path

from future.utils import with_metaclass

from snips_nlu.common.abc_utils import abstractclassmethod, classproperty
from snips_nlu.common.io_utils import temp_dir, unzip_archive
from snips_nlu.common.registrable import Registrable
from snips_nlu.common.utils import (
    json_string, check_random_state)
from snips_nlu.constants import (
    BUILTIN_ENTITY_PARSER, CUSTOM_ENTITY_PARSER, CUSTOM_ENTITY_PARSER_USAGE,
    RESOURCES, LANGUAGE, RANDOM_STATE)
from snips_nlu.entity_parser import (
    BuiltinEntityParser, CustomEntityParser, CustomEntityParserUsage)
from snips_nlu.exceptions import LoadingError
from snips_nlu.pipeline.configs import ProcessingUnitConfig
from snips_nlu.pipeline.configs.config import DefaultProcessingUnitConfig
from snips_nlu.resources import load_resources


class ProcessingUnit(with_metaclass(ABCMeta, Registrable)):
    """Abstraction of a NLU pipeline unit

    Pipeline processing units such as intent parsers, intent classifiers and
    slot fillers must implement this class.

     A :class:`ProcessingUnit` is associated with a *config_type*, which
    represents the :class:`.ProcessingUnitConfig` used to initialize it.
    """

    def __init__(self, config, **shared):
        if config is None:
            self.config = self.default_config()
        elif isinstance(config, ProcessingUnitConfig):
            self.config = config
        elif isinstance(config, dict):
            self.config = self.config_type.from_dict(config)
        else:
            raise ValueError("Unexpected config type: %s" % type(config))
        if self.config is not None:
            self.config.set_unit_name(self.unit_name)
        self.builtin_entity_parser = shared.get(BUILTIN_ENTITY_PARSER)
        self.custom_entity_parser = shared.get(CUSTOM_ENTITY_PARSER)
        self.resources = shared.get(RESOURCES)
        self.random_state = check_random_state(shared.get(RANDOM_STATE))

    @classproperty
    def config_type(cls):  # pylint:disable=no-self-argument
        return DefaultProcessingUnitConfig

    @classmethod
    def default_config(cls):
        config = cls.config_type()  # pylint:disable=no-value-for-parameter
        config.set_unit_name(cls.unit_name)
        return config

    @classproperty
    def unit_name(cls):  # pylint:disable=no-self-argument
        return ProcessingUnit.registered_name(cls)

    @classmethod
    def from_config(cls, unit_config, **shared):
        """Build a :class:`ProcessingUnit` from the provided config"""
        unit = cls.by_name(unit_config.unit_name)
        return unit(unit_config, **shared)

    @classmethod
    def load_from_path(cls, unit_path, unit_name=None, **shared):
        """Load a :class:`ProcessingUnit` from a persisted processing unit
        directory

        Args:
            unit_path (str or :class:`pathlib.Path`): path to the persisted
                processing unit
            unit_name (str, optional): Name of the processing unit to load.
                By default, the unit name is assumed to be stored in a
                "metadata.json" file located in the directory at unit_path.

        Raises:
            LoadingError: when unit_name is None and no metadata file is found
                in the processing unit directory
        """
        unit_path = Path(unit_path)
        if unit_name is None:
            metadata_path = unit_path / "metadata.json"
            if not metadata_path.exists():
                raise LoadingError(
                    "Missing metadata for processing unit at path %s"
                    % str(unit_path))
            with metadata_path.open(encoding="utf8") as f:
                metadata = json.load(f)
            unit_name = metadata["unit_name"]
        unit = cls.by_name(unit_name)
        return unit.from_path(unit_path, **shared)

    @classmethod
    def get_config(cls, unit_config):
        """Returns the :class:`.ProcessingUnitConfig` corresponding to
        *unit_config*"""
        if isinstance(unit_config, ProcessingUnitConfig):
            return unit_config
        elif isinstance(unit_config, dict):
            unit_name = unit_config["unit_name"]
            processing_unit_type = cls.by_name(unit_name)
            return processing_unit_type.config_type.from_dict(unit_config)
        elif isinstance(unit_config, (str, bytes)):
            unit_name = unit_config
            unit_config = {"unit_name": unit_name}
            processing_unit_type = cls.by_name(unit_name)
            return processing_unit_type.config_type.from_dict(unit_config)
        else:
            raise ValueError(
                "Expected `unit_config` to be an instance of "
                "ProcessingUnitConfig or dict or str but found: %s"
                % type(unit_config))

    @abstractproperty
    def fitted(self):
        """Whether or not the processing unit has already been trained"""
        pass

    def load_resources_if_needed(self, language):
        if self.resources is None or self.fitted:
            required_resources = None
            if self.config is not None:
                required_resources = self.config.get_required_resources()
            self.resources = load_resources(language, required_resources)

    def fit_builtin_entity_parser_if_needed(self, dataset):
        # We only fit a builtin entity parser when the unit has already been
        # fitted or if the parser is none.
        # In the other cases the parser is provided fitted by another unit.
        if self.builtin_entity_parser is None or self.fitted:
            self.builtin_entity_parser = BuiltinEntityParser.build(
                dataset=dataset)
        return self

    def fit_custom_entity_parser_if_needed(self, dataset):
        # We only fit a custom entity parser when the unit has already been
        # fitted or if the parser is none.
        # In the other cases the parser is provided fitted by another unit.
        required_resources = self.config.get_required_resources()
        if not required_resources or not required_resources.get(
                CUSTOM_ENTITY_PARSER_USAGE):
            # In these cases we need a custom entity parser only to do the
            # final slot resolution step, which must be done without stemming.
            parser_usage = CustomEntityParserUsage.WITHOUT_STEMS
        else:
            parser_usage = required_resources[CUSTOM_ENTITY_PARSER_USAGE]

        if self.custom_entity_parser is None or self.fitted:
            self.load_resources_if_needed(dataset[LANGUAGE])
            self.custom_entity_parser = CustomEntityParser.build(
                dataset, parser_usage, self.resources)
        return self

    def persist_metadata(self, path, **kwargs):
        metadata = {"unit_name": self.unit_name}
        metadata.update(kwargs)
        metadata_json = json_string(metadata)
        with (path / "metadata.json").open(mode="w", encoding="utf8") as f:
            f.write(metadata_json)

    @abstractmethod
    def persist(self, path):
        pass

    @abstractclassmethod
    def from_path(cls, path, **shared):
        pass

    def to_byte_array(self):
        """Serialize the :class:`ProcessingUnit` instance into a bytearray

        This method persists the processing unit in a temporary directory, zip
        the directory and return the zipped file as binary data.

        Returns:
            bytearray: the processing unit as bytearray data
        """
        cleaned_unit_name = _sanitize_unit_name(self.unit_name)
        with temp_dir() as tmp_dir:
            processing_unit_dir = tmp_dir / cleaned_unit_name
            self.persist(processing_unit_dir)
            archive_base_name = tmp_dir / cleaned_unit_name
            archive_name = archive_base_name.with_suffix(".zip")
            shutil.make_archive(
                base_name=str(archive_base_name), format="zip",
                root_dir=str(tmp_dir), base_dir=cleaned_unit_name)
            with archive_name.open(mode="rb") as f:
                processing_unit_bytes = bytearray(f.read())
        return processing_unit_bytes

    @classmethod
    def from_byte_array(cls, unit_bytes, **shared):
        """Load a :class:`ProcessingUnit` instance from a bytearray

        Args:
            unit_bytes (bytearray): A bytearray representing a zipped
                processing unit.
        """
        cleaned_unit_name = _sanitize_unit_name(cls.unit_name)
        with temp_dir() as tmp_dir:
            file_io = io.BytesIO(unit_bytes)
            unzip_archive(file_io, str(tmp_dir))
            processing_unit = cls.from_path(tmp_dir / cleaned_unit_name,
                                            **shared)
        return processing_unit


def _sanitize_unit_name(unit_name):
    return unit_name \
        .lower() \
        .strip() \
        .replace(" ", "") \
        .replace("/", "") \
        .replace("\\", "")



from __future__ import print_function, unicode_literals


def add_download_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "download", formatter_class=formatter_class,
        help="Download the language resources required by the snips-nlu "
             "library")
    subparser.add_argument(
        "resource_name", type=str,
        help="Name of the language resources to download. Can be either "
             "a shortcut, like 'en', or the full name of the resources "
             "like 'snips_nlu_en'")
    subparser.add_argument(
        "-d", "--direct", action="store_true",
        help="Force direct download. Needs resource name with version and "
             "won't perform compatibility check")
    subparser.add_argument(
        "extra_pip_args", nargs="*", type=str,
        help="Additional arguments to be passed to `pip install` when "
             "installing the language resources package")
    subparser.set_defaults(func=_download)
    return subparser


def _download(args_namespace):
    return download(args_namespace.resource_name, args_namespace.direct,
                    *args_namespace.extra_pip_args)


# inspired from
# https://github.com/explosion/spaCy/blob/master/spacy/cli/download.py

def download(resource_name, direct=False,
             *pip_args):  # pylint:disable=keyword-arg-before-vararg

    """Download compatible language resources"""
    import sys
    from snips_nlu import __about__
    from snips_nlu.cli.utils import install_remote_package

    if direct:
        components = resource_name.split("-")
        name = "".join(components[:-1])
        version = components[-1]
        url_tail = '{n}-{v}/{n}-{v}.tar.gz#egg={n}=={v}'.format(
            n=name, v=version)
        download_url = __about__.__download_url__ + '/' + url_tail
        dl = install_remote_package(download_url, pip_args)
        if dl != 0:
            sys.exit(dl)
    else:
        download_from_resource_name(resource_name, pip_args)


def download_from_resource_name(resource_name, pip_args, verbose=True):
    from snips_nlu import __about__
    from snips_nlu.cli.utils import (
        check_resources_alias, get_compatibility, get_json)

    shortcuts = get_json(__about__.__shortcuts__, "Resource shortcuts")
    check_resources_alias(resource_name, shortcuts)
    compatibility = get_compatibility()
    resource_name = resource_name.lower()
    full_resource_name = shortcuts.get(resource_name, resource_name)
    _download_and_link(resource_name, full_resource_name, compatibility,
                       pip_args, verbose)


def add_download_all_languages_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "download-all-languages", formatter_class=formatter_class,
        help="Download language resources for all languages")
    subparser.add_argument(
        "extra_pip_args", nargs="*", type=str,
        help="Additional arguments to be passed to `pip install` when "
             "installing the language resources packages")
    subparser.set_defaults(func=_download_all_languages)
    return subparser


def _download_all_languages(args_namespace):
    return download_all_languages(*args_namespace.extra_pip_args)


def download_all_languages(*pip_args):
    """Download compatible resources for all supported languages"""
    from snips_nlu_parsers import get_all_languages

    for language in get_all_languages():
        download(language, False, *pip_args)


def _download_and_link(resource_alias, resource_fullname, compatibility,
                       pip_args, verbose):
    import sys
    from snips_nlu import __about__
    from snips_nlu.cli.link import link_resources
    from snips_nlu.cli.utils import (
        PrettyPrintLevel, get_resources_version, install_remote_package,
        pretty_print)
    from snips_nlu.common.utils import get_package_path

    version = get_resources_version(resource_fullname, resource_alias,
                                    compatibility)
    url_tail = '{r}-{v}/{r}-{v}.tar.gz#egg={r}=={v}'.format(
        r=resource_fullname, v=version)
    download_url = __about__.__download_url__ + '/' + url_tail
    exit_code = install_remote_package(download_url, pip_args)
    if exit_code != 0:
        sys.exit(exit_code)
    try:
        # Get package path here because link uses
        # pip.get_installed_distributions() to check if the resource is a
        # package, which fails if the resource was just installed via
        # subprocess
        package_path = get_package_path(resource_fullname)
        link_path, resources_dir = link_resources(
            resource_fullname, resource_alias, force=True,
            resources_path=package_path)
        if verbose:
            pretty_print("%s --> %s" % (str(resources_dir), str(link_path)),
                         title="Linking successful",
                         level=PrettyPrintLevel.SUCCESS)
    except OSError as e:  # pylint:disable=bare-except
        pretty_print(
            "Creating a shortcut link for '%s' didn't work: %s"
            % (resource_alias, repr(e)),
            title="The language resources were successfully downloaded, "
                  "however linking failed.",
            level=PrettyPrintLevel.ERROR)



from __future__ import unicode_literals

from snips_nlu.common.from_dict import FromDict
from snips_nlu.constants import (
    CUSTOM_ENTITY_PARSER_USAGE, NOISE, STEMS, STOP_WORDS, WORD_CLUSTERS)
from snips_nlu.entity_parser.custom_entity_parser import (
    CustomEntityParserUsage)
from snips_nlu.pipeline.configs import Config, ProcessingUnitConfig
from snips_nlu.resources import merge_required_resources


class LogRegIntentClassifierConfig(FromDict, ProcessingUnitConfig):
    """Configuration of a :class:`.LogRegIntentClassifier`"""

    # pylint: disable=line-too-long
    def __init__(self, data_augmentation_config=None, featurizer_config=None,
                 noise_reweight_factor=1.0):
        """
        Args:
            data_augmentation_config (:class:`IntentClassifierDataAugmentationConfig`):
                    Defines the strategy of the underlying data augmentation
            featurizer_config (:class:`FeaturizerConfig`): Configuration of the
                :class:`.Featurizer` used underneath
            noise_reweight_factor (float, optional): this parameter allows to
                change the weight of the None class. By default, the class
                weights are computed using a "balanced" strategy. The
                noise_reweight_factor allows to deviate from this strategy.
        """
        if data_augmentation_config is None:
            data_augmentation_config = IntentClassifierDataAugmentationConfig()
        if featurizer_config is None:
            featurizer_config = FeaturizerConfig()
        self._data_augmentation_config = None
        self.data_augmentation_config = data_augmentation_config
        self._featurizer_config = None
        self.featurizer_config = featurizer_config
        self.noise_reweight_factor = noise_reweight_factor

    # pylint: enable=line-too-long

    @property
    def data_augmentation_config(self):
        return self._data_augmentation_config

    @data_augmentation_config.setter
    def data_augmentation_config(self, value):
        if isinstance(value, dict):
            self._data_augmentation_config = \
                IntentClassifierDataAugmentationConfig.from_dict(value)
        elif isinstance(value, IntentClassifierDataAugmentationConfig):
            self._data_augmentation_config = value
        else:
            raise TypeError("Expected instance of "
                            "IntentClassifierDataAugmentationConfig or dict"
                            "but received: %s" % type(value))

    @property
    def featurizer_config(self):
        return self._featurizer_config

    @featurizer_config.setter
    def featurizer_config(self, value):
        if isinstance(value, dict):
            self._featurizer_config = \
                FeaturizerConfig.from_dict(value)
        elif isinstance(value, FeaturizerConfig):
            self._featurizer_config = value
        else:
            raise TypeError("Expected instance of FeaturizerConfig or dict"
                            "but received: %s" % type(value))

    @property
    def unit_name(self):
        from snips_nlu.intent_classifier import LogRegIntentClassifier
        return LogRegIntentClassifier.unit_name

    def get_required_resources(self):
        resources = self.data_augmentation_config.get_required_resources()
        resources = merge_required_resources(
            resources, self.featurizer_config.get_required_resources())
        return resources

    def to_dict(self):
        return {
            "unit_name": self.unit_name,
            "data_augmentation_config":
                self.data_augmentation_config.to_dict(),
            "featurizer_config": self.featurizer_config.to_dict(),
            "noise_reweight_factor": self.noise_reweight_factor,
        }


class IntentClassifierDataAugmentationConfig(FromDict, Config):
    """Configuration used by a :class:`.LogRegIntentClassifier` which defines
        how to augment data to improve the training of the classifier"""

    def __init__(self, min_utterances=20, noise_factor=5,
                 add_builtin_entities_examples=True, unknown_word_prob=0,
                 unknown_words_replacement_string=None,
                 max_unknown_words=None):
        """
        Args:
            min_utterances (int, optional): The minimum number of utterances to
                automatically generate for each intent, based on the existing
                utterances. Default is 20.
            noise_factor (int, optional): Defines the size of the noise to
                generate to train the implicit *None* intent, as a multiplier
                of the average size of the other intents. Default is 5.
            add_builtin_entities_examples (bool, optional): If True, some
                builtin entity examples will be automatically added to the
                training data. Default is True.
        """
        self.min_utterances = min_utterances
        self.noise_factor = noise_factor
        self.add_builtin_entities_examples = add_builtin_entities_examples
        self.unknown_word_prob = unknown_word_prob
        self.unknown_words_replacement_string = \
            unknown_words_replacement_string
        if max_unknown_words is not None and max_unknown_words < 0:
            raise ValueError("max_unknown_words must be None or >= 0")
        self.max_unknown_words = max_unknown_words
        if unknown_word_prob > 0 and unknown_words_replacement_string is None:
            raise ValueError("unknown_word_prob is positive (%s) but the "
                             "replacement string is None" % unknown_word_prob)

    @staticmethod
    def get_required_resources():
        return {
            NOISE: True,
            STOP_WORDS: True
        }

    def to_dict(self):
        return {
            "min_utterances": self.min_utterances,
            "noise_factor": self.noise_factor,
            "add_builtin_entities_examples":
                self.add_builtin_entities_examples,
            "unknown_word_prob": self.unknown_word_prob,
            "unknown_words_replacement_string":
                self.unknown_words_replacement_string,
            "max_unknown_words": self.max_unknown_words
        }


class FeaturizerConfig(FromDict, ProcessingUnitConfig):
    """Configuration of a :class:`.Featurizer` object"""

    # pylint: disable=line-too-long
    def __init__(self, tfidf_vectorizer_config=None,
                 cooccurrence_vectorizer_config=None,
                 pvalue_threshold=0.4,
                 added_cooccurrence_feature_ratio=0):
        """
        Args:
            tfidf_vectorizer_config (:class:`.TfidfVectorizerConfig`, optional):
                empty configuration of the featurizer's
                :attr:`tfidf_vectorizer`
            cooccurrence_vectorizer_config: (:class:`.CooccurrenceVectorizerConfig`, optional):
                configuration of the featurizer's
                :attr:`cooccurrence_vectorizer`
            pvalue_threshold (float): after fitting the training set to
                extract tfidf features, a univariate feature selection is
                applied. Features are tested for independence using a Chi-2
                test, under the null hypothesis that each feature should be
                equally present in each class. Only features having a p-value
                lower than the threshold are kept
            added_cooccurrence_feature_ratio (float, optional): proportion of
                cooccurrence features to add with respect to the number of
                tfidf features. For instance with a ratio of 0.5, if 100 tfidf
                features are remaining after feature selection, a maximum of 50
                cooccurrence features will be added
        """
        self.pvalue_threshold = pvalue_threshold
        self.added_cooccurrence_feature_ratio = \
            added_cooccurrence_feature_ratio

        if tfidf_vectorizer_config is None:
            tfidf_vectorizer_config = TfidfVectorizerConfig()
        elif isinstance(tfidf_vectorizer_config, dict):
            tfidf_vectorizer_config = TfidfVectorizerConfig.from_dict(
                tfidf_vectorizer_config)
        self.tfidf_vectorizer_config = tfidf_vectorizer_config

        if cooccurrence_vectorizer_config is None:
            cooccurrence_vectorizer_config = CooccurrenceVectorizerConfig()
        elif isinstance(cooccurrence_vectorizer_config, dict):
            cooccurrence_vectorizer_config = CooccurrenceVectorizerConfig \
                .from_dict(cooccurrence_vectorizer_config)
        self.cooccurrence_vectorizer_config = cooccurrence_vectorizer_config

    # pylint: enable=line-too-long

    @property
    def unit_name(self):
        from snips_nlu.intent_classifier import Featurizer
        return Featurizer.unit_name

    def get_required_resources(self):
        required_resources = self.tfidf_vectorizer_config \
            .get_required_resources()
        if self.cooccurrence_vectorizer_config:
            required_resources = merge_required_resources(
                required_resources,
                self.cooccurrence_vectorizer_config.get_required_resources())
        return required_resources

    def to_dict(self):
        return {
            "unit_name": self.unit_name,
            "pvalue_threshold": self.pvalue_threshold,
            "added_cooccurrence_feature_ratio":
                self.added_cooccurrence_feature_ratio,
            "tfidf_vectorizer_config": self.tfidf_vectorizer_config.to_dict(),
            "cooccurrence_vectorizer_config":
                self.cooccurrence_vectorizer_config.to_dict(),
        }


class TfidfVectorizerConfig(FromDict, ProcessingUnitConfig):
    """Configuration of a :class:`.TfidfVectorizerConfig` object"""

    def __init__(self, word_clusters_name=None, use_stemming=False):
        """
        Args:
            word_clusters_name (str, optional): if a word cluster name is
                provided then the featurizer will use the word clusters IDs
                detected in the utterances and add them to the utterance text
                before computing the tfidf. Default to None
            use_stemming (bool, optional): use stemming before computing the
                tfdif. Defaults to False (no stemming used)
        """
        self.word_clusters_name = word_clusters_name
        self.use_stemming = use_stemming

    @property
    def unit_name(self):
        from snips_nlu.intent_classifier import TfidfVectorizer
        return TfidfVectorizer.unit_name

    def get_required_resources(self):
        resources = {STEMS: True if self.use_stemming else False}
        if self.word_clusters_name:
            resources[WORD_CLUSTERS] = {self.word_clusters_name}
        return resources

    def to_dict(self):
        return {
            "unit_name": self.unit_name,
            "word_clusters_name": self.word_clusters_name,
            "use_stemming": self.use_stemming
        }


class CooccurrenceVectorizerConfig(FromDict, ProcessingUnitConfig):
    """Configuration of a :class:`.CooccurrenceVectorizer` object"""

    def __init__(self, window_size=None, unknown_words_replacement_string=None,
                 filter_stop_words=True, keep_order=True):
        """
        Args:
            window_size (int, optional): if provided, word cooccurrences will
                be taken into account only in a context window of size
                :attr:`window_size`. If the window size is 3 then given a word
                w[i], the vectorizer will only extract the following pairs:
                (w[i], w[i + 1]), (w[i], w[i + 2]) and (w[i], w[i + 3]).
                Defaults to None, which means that we consider all words
            unknown_words_replacement_string (str, optional)
            filter_stop_words (bool, optional): if True, stop words are ignored
                when computing cooccurrences
            keep_order (bool, optional): if True then cooccurrence are computed
                taking the words order into account, which means the pairs
                (w1, w2) and (w2, w1) will count as two separate features.
                Defaults to `True`.
        """
        self.window_size = window_size
        self.unknown_words_replacement_string = \
            unknown_words_replacement_string
        self.filter_stop_words = filter_stop_words
        self.keep_order = keep_order

    @property
    def unit_name(self):
        from snips_nlu.intent_classifier import CooccurrenceVectorizer
        return CooccurrenceVectorizer.unit_name

    def get_required_resources(self):
        return {
            STOP_WORDS: self.filter_stop_words,
            # We require the parser to be trained without stems because we
            # don't normalize and stem when processing in the
            # CooccurrenceVectorizer (in order to run the builtin and
            # custom parser on the same unormalized input).
            # Requiring no stems ensures we'll be able to parse the unstemmed
            # input
            CUSTOM_ENTITY_PARSER_USAGE: CustomEntityParserUsage.WITHOUT_STEMS
        }

    def to_dict(self):
        return {
            "unit_name": self.unit_name,
            "unknown_words_replacement_string":
                self.unknown_words_replacement_string,
            "window_size": self.window_size,
            "filter_stop_words": self.filter_stop_words,
            "keep_order": self.keep_order
        }



# coding=utf-8
from __future__ import unicode_literals

import json
import operator
from copy import deepcopy
from pathlib import Path

from future.utils import iteritems, viewvalues

from snips_nlu.common.utils import json_string
from snips_nlu.constants import (
    END, ENTITIES, LANGUAGE, MATCHING_STRICTNESS, START, UTTERANCES,
    LICENSE_INFO)
from snips_nlu.entity_parser.builtin_entity_parser import is_builtin_entity
from snips_nlu.entity_parser.custom_entity_parser_usage import (
    CustomEntityParserUsage)
from snips_nlu.entity_parser.entity_parser import EntityParser
from snips_nlu.preprocessing import stem, tokenize, tokenize_light
from snips_nlu.result import parsed_entity

STOPWORDS_FRACTION = 1e-3


class CustomEntityParser(EntityParser):
    def __init__(self, parser, language, parser_usage):
        super(CustomEntityParser, self).__init__()
        self._parser = parser
        self.language = language
        self.parser_usage = parser_usage

    def _parse(self, text, scope=None):
        tokens = tokenize(text, self.language)
        shifts = _compute_char_shifts(tokens)
        cleaned_text = " ".join(token.value for token in tokens)

        entities = self._parser.parse(cleaned_text, scope)
        result = []
        for entity in entities:
            start = entity["range"]["start"]
            start -= shifts[start]
            end = entity["range"]["end"]
            end -= shifts[end - 1]
            entity_range = {START: start, END: end}
            ent = parsed_entity(
                entity_kind=entity["entity_identifier"],
                entity_value=entity["value"],
                entity_resolved_value=entity["resolved_value"],
                entity_range=entity_range
            )
            result.append(ent)
        return result

    def persist(self, path):
        path = Path(path)
        path.mkdir()
        parser_directory = "parser"
        metadata = {
            "language": self.language,
            "parser_usage": self.parser_usage.value,
            "parser_directory": parser_directory
        }
        with (path / "metadata.json").open(mode="w", encoding="utf8") as f:
            f.write(json_string(metadata))
        self._parser.persist(path / parser_directory)

    @classmethod
    def from_path(cls, path):
        from snips_nlu_parsers import GazetteerEntityParser

        path = Path(path)
        with (path / "metadata.json").open(encoding="utf8") as f:
            metadata = json.load(f)
        language = metadata["language"]
        parser_usage = CustomEntityParserUsage(metadata["parser_usage"])
        parser_path = path / metadata["parser_directory"]
        parser = GazetteerEntityParser.from_path(parser_path)
        return cls(parser, language, parser_usage)

    @classmethod
    def build(cls, dataset, parser_usage, resources):
        from snips_nlu_parsers import GazetteerEntityParser
        from snips_nlu.dataset import validate_and_format_dataset

        dataset = validate_and_format_dataset(dataset)
        language = dataset[LANGUAGE]
        custom_entities = {
            entity_name: deepcopy(entity)
            for entity_name, entity in iteritems(dataset[ENTITIES])
            if not is_builtin_entity(entity_name)
        }
        if parser_usage == CustomEntityParserUsage.WITH_AND_WITHOUT_STEMS:
            for ent in viewvalues(custom_entities):
                stemmed_utterances = _stem_entity_utterances(
                    ent[UTTERANCES], language, resources)
                ent[UTTERANCES] = _merge_entity_utterances(
                    ent[UTTERANCES], stemmed_utterances)
        elif parser_usage == CustomEntityParserUsage.WITH_STEMS:
            for ent in viewvalues(custom_entities):
                ent[UTTERANCES] = _stem_entity_utterances(
                    ent[UTTERANCES], language, resources)
        elif parser_usage is None:
            raise ValueError("A parser usage must be defined in order to fit "
                             "a CustomEntityParser")
        configuration = _create_custom_entity_parser_configuration(
            custom_entities,
            language=dataset[LANGUAGE],
            stopwords_fraction=STOPWORDS_FRACTION,
        )
        parser = GazetteerEntityParser.build(configuration)
        return cls(parser, language, parser_usage)


def _stem_entity_utterances(entity_utterances, language, resources):
    values = dict()
    # Sort by resolved value, so that values conflict in a deterministic way
    for raw_value, resolved_value in sorted(
            iteritems(entity_utterances), key=operator.itemgetter(1)):
        stemmed_value = stem(raw_value, language, resources)
        if stemmed_value not in values:
            values[stemmed_value] = resolved_value
    return values


def _merge_entity_utterances(raw_utterances, stemmed_utterances):
    # Sort by resolved value, so that values conflict in a deterministic way
    for raw_stemmed_value, resolved_value in sorted(
            iteritems(stemmed_utterances), key=operator.itemgetter(1)):
        if raw_stemmed_value not in raw_utterances:
            raw_utterances[raw_stemmed_value] = resolved_value
    return raw_utterances


def _create_custom_entity_parser_configuration(
        entities, stopwords_fraction, language):
    """Dynamically creates the gazetteer parser configuration.

    Args:
        entities (dict): entity for the dataset
        stopwords_fraction (float): fraction of the vocabulary of
            the entity values that will be considered as stop words (
            the top n_vocabulary * stopwords_fraction most frequent words will
            be considered stop words)
        language (str): language of the entities

    Returns: the parser configuration as dictionary
    """

    if not 0 < stopwords_fraction < 1:
        raise ValueError("stopwords_fraction must be in ]0.0, 1.0[")

    parser_configurations = []
    for entity_name, entity in sorted(iteritems(entities)):
        vocabulary = set(
            t for raw_value in entity[UTTERANCES]
            for t in tokenize_light(raw_value, language)
        )
        num_stopwords = int(stopwords_fraction * len(vocabulary))
        config = {
            "entity_identifier": entity_name,
            "entity_parser": {
                "threshold": entity[MATCHING_STRICTNESS],
                "n_gazetteer_stop_words": num_stopwords,
                "gazetteer": [
                    {
                        "raw_value": k,
                        "resolved_value": v
                    } for k, v in sorted(iteritems(entity[UTTERANCES]))
                ]
            }
        }
        if LICENSE_INFO in entity:
            config["entity_parser"][LICENSE_INFO] = entity[LICENSE_INFO]
        parser_configurations.append(config)

    configuration = {
        "entity_parsers": parser_configurations
    }

    return configuration


def _compute_char_shifts(tokens):
    """Compute the shifts in characters that occur when comparing the
    tokens string with the string consisting of all tokens separated with a
    space

    For instance, if "hello?world" is tokenized in ["hello", "?", "world"],
    then the character shifts between "hello?world" and "hello ? world" are
    [0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 2, 2, 2]
    """
    characters_shifts = []
    if not tokens:
        return characters_shifts

    current_shift = 0
    for token_index, token in enumerate(tokens):
        if token_index == 0:
            previous_token_end = 0
            previous_space_len = 0
        else:
            previous_token_end = tokens[token_index - 1].end
            previous_space_len = 1
        offset = (token.start - previous_token_end) - previous_space_len
        current_shift -= offset
        token_len = token.end - token.start
        index_shift = token_len + previous_space_len
        characters_shifts += [current_shift for _ in range(index_shift)]
    return characters_shifts



from snips_nlu.constants import INTENTS, UTTERANCES, DATA, SLOT_NAME, ENTITY
from snips_nlu.exceptions import DatasetFormatError


def type_error(expected_type, found_type, object_label=None):
    if object_label is None:
        raise DatasetFormatError("Invalid type: expected %s but found %s"
                                 % (expected_type, found_type))
    raise DatasetFormatError("Invalid type for '%s': expected %s but found %s"
                             % (object_label, expected_type, found_type))


def validate_type(obj, expected_type, object_label=None):
    if not isinstance(obj, expected_type):
        type_error(expected_type, type(obj), object_label)


def missing_key_error(key, object_label=None):
    if object_label is None:
        raise DatasetFormatError("Missing key: '%s'" % key)
    raise DatasetFormatError("Expected %s to have key: '%s'"
                             % (object_label, key))


def validate_key(obj, key, object_label=None):
    if key not in obj:
        missing_key_error(key, object_label)


def validate_keys(obj, keys, object_label=None):
    for key in keys:
        validate_key(obj, key, object_label)


def get_slot_name_mapping(dataset, intent):
    """Returns a dict which maps slot names to entities for the provided intent
    """
    slot_name_mapping = dict()
    for utterance in dataset[INTENTS][intent][UTTERANCES]:
        for chunk in utterance[DATA]:
            if SLOT_NAME in chunk:
                slot_name_mapping[chunk[SLOT_NAME]] = chunk[ENTITY]
    return slot_name_mapping


def get_slot_name_mappings(dataset):
    """Returns a dict which maps intents to their slot name mapping"""
    return {intent: get_slot_name_mapping(dataset, intent)
            for intent in dataset[INTENTS]}



from collections import OrderedDict


class LimitedSizeDict(OrderedDict):
    def __init__(self, *args, **kwds):
        if "size_limit" not in kwds:
            raise ValueError("'size_limit' must be passed as a keyword "
                             "argument")
        self.size_limit = kwds.pop("size_limit")
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        if len(args) == 1 and len(args[0]) + len(kwds) > self.size_limit:
            raise ValueError("Tried to initialize LimitedSizedDict with more "
                             "value than permitted with 'limit_size'")
        super(LimitedSizeDict, self).__init__(*args, **kwds)

    def __setitem__(self, key, value, dict_setitem=OrderedDict.__setitem__):
        dict_setitem(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)

    def __eq__(self, other):
        if self.size_limit != other.size_limit:
            return False
        return super(LimitedSizeDict, self).__eq__(other)


class UnupdatableDict(dict):
    def __setitem__(self, key, value):
        if key in self:
            raise KeyError("Can't update key '%s'" % key)
        super(UnupdatableDict, self).__setitem__(key, value)



from __future__ import unicode_literals, print_function


def add_parse_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "parse", formatter_class=formatter_class,
        help="Load a trained NLU engine and perform parsing")
    subparser.add_argument("training_path", type=str,
                           help="Path to a trained engine")
    subparser.add_argument("-q", "--query", type=str,
                           help="Query to parse. If provided, it disables the "
                                "interactive behavior.")
    subparser.add_argument("-v", "--verbosity", action="count", default=0,
                           help="Increase output verbosity")
    subparser.add_argument("-f", "--intents-filter", type=str,
                           help="Intents filter as a comma-separated list")
    subparser.set_defaults(func=_parse)
    return subparser


def _parse(args_namespace):
    return parse(args_namespace.training_path, args_namespace.query,
                 args_namespace.verbosity, args_namespace.intents_filter)


def parse(training_path, query, verbose=False, intents_filter=None):
    """Load a trained NLU engine and play with its parsing API interactively"""
    import csv
    import logging
    from builtins import input, str
    from snips_nlu import SnipsNLUEngine
    from snips_nlu.cli.utils import set_nlu_logger

    if verbose == 1:
        set_nlu_logger(logging.INFO)
    elif verbose >= 2:
        set_nlu_logger(logging.DEBUG)
    if intents_filter:
        # use csv in order to properly handle commas and other special
        # characters in intent names
        intents_filter = next(csv.reader([intents_filter]))
    else:
        intents_filter = None

    engine = SnipsNLUEngine.from_path(training_path)

    if query:
        print_parsing_result(engine, query, intents_filter)
        return

    while True:
        query = input("Enter a query (type 'q' to quit): ").strip()
        if not isinstance(query, str):
            query = query.decode("utf-8")
        if query == "q":
            break
        print_parsing_result(engine, query, intents_filter)


def print_parsing_result(engine, query, intents_filter):
    from snips_nlu.common.utils import unicode_string, json_string

    query = unicode_string(query)
    json_dump = json_string(engine.parse(query, intents_filter),
                            sort_keys=True, indent=2)
    print(json_dump)



from __future__ import print_function, unicode_literals


def add_link_parser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "link", formatter_class=formatter_class,
        help="Manually link downloaded resources")
    subparser.add_argument("origin", type=str,
                           help="Package name or local path to model")
    subparser.add_argument("link_name", type=str,
                           help="Name of the symbolic link which will be "
                                "created")
    subparser.add_argument("-f", "--force", action="store_true",
                           help="Force overwriting of existing link")
    subparser.set_defaults(func=_link)
    return subparser


def _link(args_namespace):
    return link(args_namespace.origin, args_namespace.link_name,
                args_namespace.force)


# inspired from
# https://github.com/explosion/spaCy/blob/master/spacy/cli/link.py

def link(origin, link_name, force=False, resources_path=None):
    """
    Create a symlink for language resources within the snips_nlu/data
    directory. Accepts either the name of a pip package, or the local path to
    the resources data directory.

    Linking resources allows loading them via
    snips_nlu.load_resources(link_name).
    """
    from snips_nlu.cli.utils import PrettyPrintLevel, pretty_print

    link_path, resources_dir = link_resources(origin, link_name, force,
                                              resources_path)
    pretty_print("%s --> %s" % (str(resources_dir), str(link_path)),
                 title="Linking successful", level=PrettyPrintLevel.SUCCESS)


def link_resources(origin, link_name, force, resources_path):
    from pathlib import Path
    from snips_nlu.cli.compatibility import create_symlink
    from snips_nlu.common.utils import get_package_path, is_package
    from snips_nlu.constants import DATA_PATH
    from snips_nlu.resources import get_resources_sub_directory

    if is_package(origin):
        resources_path = get_package_path(origin)
    else:
        resources_path = Path(origin) if resources_path is None \
            else Path(resources_path)
    if not resources_path.exists():
        raise OSError("%s not found" % str(resources_path))
    link_path = DATA_PATH / str(link_name)
    if link_path.is_symlink() and not force:
        raise OSError("Symlink already exists: %s" % str(link_path))
    elif link_path.is_symlink():
        link_path.unlink()
    elif link_path.exists():
        raise OSError("Symlink cannot be overwritten: %s" % str(link_path))
    resources_sub_dir = get_resources_sub_directory(resources_path)
    create_symlink(link_path, resources_sub_dir)
    return link_path, resources_sub_dir



from __future__ import print_function, unicode_literals


def add_generate_dataset_subparser(subparsers, formatter_class):
    subparser = subparsers.add_parser(
        "generate-dataset", formatter_class=formatter_class,
        help="Generate a json dataset from intents and entities yaml files")
    subparser.add_argument("language", type=str,
                           help="Language of the dataset")
    subparser.add_argument("files", nargs="+", type=str,
                           help="List of intent and entity yaml files")
    subparser.set_defaults(func=_generate_dataset)
    return subparser


def _generate_dataset(args_namespace):
    return generate_dataset(args_namespace.language, *args_namespace.files)


def generate_dataset(language, *yaml_files):
    """Creates a Snips NLU dataset from YAML definition files

    Check :meth:`.Intent.from_yaml` and :meth:`.Entity.from_yaml` for the
    format of the YAML files.

    Args:
        language (str): language of the dataset (iso code)
        *yaml_files: list of intent and entity definition files in YAML format.

    Returns:
        None. The json dataset output is printed out on stdout.
    """
    from snips_nlu.dataset import Dataset
    from snips_nlu.common.utils import unicode_string, json_string

    language = unicode_string(language)
    dataset = Dataset.from_yaml_files(language, list(yaml_files))
    print(json_string(dataset.json, indent=2, sort_keys=True))
